interface Ok<T> {
    ok: true;
    /**
     * The output value.
     */
    value: T;
}
interface Err {
    ok: false;
    /**
     * The list of issues encountered during parsing.
     */
    issues: Issue[];
}
/**
 * A callback that takes a value and returns `undefined` if it satisfies the requirements. If a value doesn't satisfy
 * the check requirements then issues can be returned or a {@linkcode ValidationError} can be thrown.
 *
 * @type T The value type.
 */
type CheckCallback<T = any> = (value: T, options: Readonly<ParseOptions>) => Partial<Issue>[] | Partial<Issue> | null | undefined | void;
/**
 * The shape output value check.
 */
interface Check {
    /**
     * The callback that validates the shape output and returns the list of issues or throws a {@linkcode Validation} error.
     */
    readonly callback: CheckCallback;
    /**
     * `true` if the {@linkcode callback} is invoked even if previous processors failed, or `false` otherwise.
     */
    readonly unsafe: boolean;
    /**
     * The optional parameter used by the {@linkcode callback}.
     */
    readonly param: any;
}
/**
 * Options that are applicable for the custom checks added via {@linkcode Shape.check}.
 */
interface CheckOptions {
    /**
     * The unique key of the check in scope of the shape.
     */
    key?: unknown;
    /**
     * If `true` then the check would be executed even if the preceding check failed, otherwise check is
     * ignored.
     */
    unsafe?: boolean;
    /**
     * An optional param associated with the check.
     */
    param?: any;
}
/**
 * A validation issue raised during input parsing.
 */
interface Issue {
    /**
     * The code of the validation issue.
     */
    code: any;
    /**
     * The object path where an issue has occurred.
     */
    path: any[];
    /**
     * The value that caused an issue to occur.
     */
    input: any;
    /**
     * A message associated with an issue. Built-in messages are strings but custom messages can have an arbitrary type.
     */
    message: any;
    /**
     * An additional param that is specific for a particular {@linkcode code}.
     */
    param: any;
    /**
     * An arbitrary metadata that can be used for message formatting.
     */
    meta: any;
}
/**
 * The message callback or a string.
 *
 * @param param The check param or `undefined` if there's no param.
 * @param code The issue code.
 * @param input The input value that the shape was trying to parse.
 * @param meta The metadata passed as a check option during shape declaration.
 * @param options The parsing options.
 * @returns Any value that should be used as an issue message.
 */
type MessageCallback = (param: any, code: any, input: any, meta: any, options: Readonly<ParseOptions>) => any;
/**
 * A callback that returns an issue message or a message string.
 */
type Message = MessageCallback | string;
/**
 * Options that are applicable for the type constraint.
 */
interface TypeConstraintOptions {
    /**
     * The custom issue message.
     */
    message?: Message | Literal;
    /**
     * An arbitrary metadata that is added to an issue.
     */
    meta?: any;
}
/**
 * Options that are applicable for the built-in type-specific constraints.
 */
interface ConstraintOptions {
    /**
     * The custom issue message.
     */
    message?: Message | Literal;
    /**
     * An arbitrary metadata that is added to an issue.
     */
    meta?: any;
    /**
     * If `true` then the constraint would be executed even if the preceding check failed, otherwise the constraint is
     * ignored.
     */
    unsafe?: boolean;
}
/**
 * Options applied during parsing.
 */
interface ParseOptions {
    /**
     * If `true` then all issues are collected during parsing, otherwise parsing is aborted after the first issue is
     * encountered.
     *
     * @default false
     */
    verbose?: boolean;
    /**
     * If `true` then all shapes would try to coerce input to a required type.
     *
     * @default false
     */
    coerced?: boolean;
    /**
     * The custom context. Use it to pass custom params to check and transform callbacks.
     */
    context?: any;
}
type Literal = object | string | number | bigint | boolean | symbol | null | undefined;

/**
 * The unique symbol that is used for type branding.
 */
declare const BRAND: unique symbol;
/**
 * An arbitrary shape.
 */
type AnyShape = Shape | Shape<never>;
/**
 * A shape that replaces an input value with an output value.
 *
 * @template S The base shape.
 * @template A The input value to replace.
 * @template B The output value.
 */
type OpaqueReplaceShape<S extends AnyShape, A, B = A> = Shape<S['input'] | A, Exclude<S['output'], A> | B>;
/**
 * Input value is passed directly to the output without any checks.
 *
 * @template S The base shape.
 * @template T The included value.
 */
type OpaqueIncludeShape<S extends AnyShape, T> = Shape<S['input'] | T, S['output'] | T>;
/**
 * Excludes a value from both input and output.
 *
 * @template S The base shape.
 * @template T The excluded value.
 */
type OpaqueExcludeShape<S extends AnyShape, T> = Shape<Exclude<S['input'], T>, Exclude<S['output'], T>>;
/**
 * The detected runtime input value type.
 */
type ValueType = 'object' | 'array' | 'function' | 'string' | 'symbol' | 'number' | 'bigint' | 'boolean' | 'date' | 'null' | 'undefined' | 'any' | 'never';
/**
 * The result of shape application. This is the part of the internal API required for creating custom shapes.
 */
type ApplyResult<T = any> = Ok<T> | Issue[] | null;
/**
 * The callback to which shape checks are compiled, see {@linkcode Shape._applyChecks}.
 */
type ApplyChecksCallback = (output: any, issues: Issue[] | null, options: ParseOptions) => Issue[] | null;
interface ReadonlyDict<T = any> {
    readonly [key: string]: T;
}
/**
 * The baseline shape implementation.
 *
 * @template I The input value.
 * @template O The output value.
 */
declare class Shape<I = any, O = I> {
    /**
     * The human-readable shape description.
     */
    description: string;
    /**
     * The map from a check key to {@linkcode Check} object, or `undefined` if this shape has no associated checks.
     */
    protected _checkMap: ReadonlyMap<unknown, Check> | undefined;
    /**
     * A callback that applies checks to the given value.
     */
    protected _applyChecks: ApplyChecksCallback | null;
    /**
     * `true` if some checks from {@linkcode _checkMap} were marked as unsafe, `false` otherwise.
     */
    protected _unsafe: boolean;
    /**
     * Returns a sub-shape that describes a value associated with the given property name, or `null` if there's no such
     * sub-shape.
     *
     * @param key The key for which the sub-shape must be retrieved.
     * @returns The sub-shape or `null` if there's no such key in the shape.
     */
    at(key: unknown): AnyShape | null;
    /**
     * Adds a human-readable description text to the shape.
     *
     * @param text The description text.
     * @returns The clone of the shape with the description added.
     */
    describe(text: string): this;
    /**
     * Adds the check that is applied to the shape output.
     *
     * If the {@linkcode CheckOptions.key} is defined and there's already a check with the same key then it is replaced.
     * If the key is `undefined` then the `cb` identity is used as a key.
     *
     * If check callback returns an empty array, it is considered that no issues have occurred.
     *
     * @param cb The callback that checks the shape output.
     * @param options The check options.
     * @returns The clone of the shape with the check added.
     */
    check(cb: CheckCallback<O>, options?: CheckOptions): this;
    /**
     * Returns the {@linkcode Check} by its key.
     *
     * @param key The check key.
     * @returns The check or `undefined` if there's no check associated with the key.
     */
    getCheck(key: unknown): Check | undefined;
    /**
     * Deletes the check from the shape.
     *
     * @param key The check key.
     * @returns The clone of the shape if check was deleted or this shape if there are no matching check.
     */
    deleteCheck(key: unknown): this;
    /**
     * Refines the shape output type with the [narrowing predicate](https://www.typescriptlang.org/docs/handbook/2/narrowing.html).
     *
     * @param cb The predicate that returns `true` if value conforms the required type, or `false` otherwise.
     * @param options The constraint options or an issue message.
     * @returns The shape with the narrowed output.
     * @template T The narrowed output value.
     */
    refine<T extends O>(
    /**
     * @param output The shape output value.
     */
    cb: (output: O) => output is T, options?: ConstraintOptions | Message): Shape<I, T>;
    /**
     * Checks that the output value conforms the predicate.
     *
     * @param cb The predicate that returns truthy result if value is valid, or returns falsy result otherwise.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    refine(
    /**
     * @param output The shape output value.
     */
    cb: (output: O) => boolean, options?: ConstraintOptions | Message): this;
    /**
     * Pipes the shape output to another shape.
     *
     * @param shape The shape that validates the output if this shape.
     * @returns The {@linkcode PipeShape} instance.
     * @template T The output value.
     */
    to<T>(shape: Shape<O, T>): Shape<I, T>;
    /**
     * Synchronously transforms the output value of the shape with a transformer callback.
     *
     * @param cb The callback that transforms the shape output value.
     * @returns The {@linkcode TransformShape} instance.
     * @template T The output value.
     */
    transform<T>(
    /**
     * @param output The shape output value.
     * @param options Parsing options.
     * @return The transformed value.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    cb: (output: O, options: Readonly<ParseOptions>) => T): Shape<I, T>;
    /**
     * Asynchronously transforms the output value of the shape with a transformer callback.
     *
     * @param cb The callback that transforms the shape output value.
     * @returns The {@linkcode TransformShape} instance.
     * @template T The transformed value.
     */
    transformAsync<T>(
    /**
     * @param output The shape output value.
     * @param options Parsing options.
     * @return The transformed value.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    cb: (output: O, options: Readonly<ParseOptions>) => Promise<T>): Shape<I, T>;
    /**
     * Adds a brand to the output type of the shape.
     *
     * @returns An opaque shape with branded output type.
     * @template T The unique brand marker. By default, this shape is used as a marker.
     */
    brand<T = this>(): Shape<I, O & {
        [BRAND]: T;
    }>;
    /**
     * Replaces an input value with an output value.
     *
     * @param inputValue The input value to replace.
     * @param outputValue The output value that is returned if an `inputValue` is received.
     * @returns The {@linkcode ReplaceShape} instance.
     * @template A The input value to replace.
     * @template B The output value.
     */
    replace<A extends Literal, B extends Literal>(inputValue: A, outputValue: B): OpaqueReplaceShape<this, A, B>;
    /**
     * Input value is passed directly to the output without any checks.
     *
     * @param value The included value.
     * @param options The constraint options or an issue message.
     * @returns The {@linkcode ReplaceShape} instance.
     * @template T The included value.
     */
    include<T extends Literal>(value: T, options?: TypeConstraintOptions | Message): OpaqueIncludeShape<this, T>;
    /**
     * Excludes value from both input and output.
     *
     * @param value The excluded value.
     * @param options The constraint options or an issue message.
     * @returns The {@linkcode ExcludeShape} instance.
     * @template T The excluded value.
     */
    exclude<T extends Literal>(value: T, options?: TypeConstraintOptions | Message): OpaqueExcludeShape<this, T>;
    /**
     * Replaces `undefined` input value with an `undefined` output value.
     *
     * @returns The {@linkcode ReplaceShape} instance.
     */
    optional(): OpaqueReplaceShape<this, undefined>;
    /**
     * Replaces `undefined` input value with a default output value.
     *
     * @param defaultValue The value that should be used if an input value is `undefined`.
     * @returns The {@linkcode ReplaceShape} instance.
     */
    optional<T extends Literal>(defaultValue: T): OpaqueReplaceShape<this, undefined, T>;
    /**
     * Replaces `null` input value with an `null` output value.
     *
     * @returns The {@linkcode ReplaceShape} instance.
     */
    nullable(): OpaqueReplaceShape<this, null>;
    /**
     * Replaces `null` input value with a default output value.
     *
     * @param defaultValue The value that should be used if an input value is `null`.
     * @returns The {@linkcode ReplaceShape} instance.
     */
    nullable<T extends Literal>(defaultValue: T): OpaqueReplaceShape<this, null, T>;
    /**
     * Passes `null` and `undefined` input values directly to the output without parsing.
     *
     * @returns The {@linkcode ReplaceShape} instance.
     */
    nullish(): OpaqueReplaceShape<this, null | undefined>;
    /**
     * Replaces `null` and `undefined` input value with a default output value.
     *
     * @param defaultValue The value that should be used if an input value is `undefined` or `null`.
     * @returns The {@linkcode ReplaceShape} instance.
     */
    nullish<T extends Literal>(defaultValue?: T): OpaqueReplaceShape<this, null | undefined, T>;
    /**
     * Prevents an input and output from being `undefined`.
     *
     * @param options The constraint options or an issue message.
     * @returns The {@linkcode ExcludeShape} instance.
     */
    nonOptional(options?: TypeConstraintOptions | Message): OpaqueExcludeShape<this, undefined>;
    /**
     * Returns `undefined` if parsing fails.
     *
     * @returns The {@linkcode CatchShape} instance.
     */
    catch(): Shape<I, O | undefined>;
    /**
     * Returns the fallback value if parsing fails.
     *
     * @param fallback The value or a callback that returns a value that is returned if parsing has failed.
     * @returns The {@linkcode CatchShape} instance.
     */
    catch<T extends Literal>(fallback: T | (() => T)): Shape<I, O | T>;
    /**
     * Must return `true` if the shape must be used in async context only, otherwise the shape can be used in both sync
     * and async contexts. Override this method to implement a custom shape.
     */
    protected _requiresAsync(): boolean;
    /**
     * Returns the list of runtime value types that can be processed by the shape. Used for various optimizations.
     */
    protected _getInputTypes(): ValueType[];
    /**
     * Returns the list of all input values that are known beforehand.
     */
    protected _getInputValues(): unknown[];
    /**
     * Synchronously parses the input.
     *
     * Override this method to implement a custom shape.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @returns `null` if input matches the output, {@linkcode Ok} that wraps the output, or an array of captured issues.
     */
    protected _apply(input: unknown, options: ParseOptions): ApplyResult<O>;
    /**
     * Asynchronously parses the input.
     *
     * Override this method to implement a custom shape that requires an async execution context.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @returns `null` if input matches the output, {@linkcode Ok} that wraps the output, or an array of captured issues.
     */
    protected _applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<O>>;
    /**
     * Returns a shape clone with new set of checks.
     *
     * @param checkMap The map from a check key to a corresponding check.
     * @returns The clone of the shape.
     */
    protected _replaceChecks(checkMap: ReadonlyMap<unknown, Check>): this;
}
interface Shape<I, O> {
    /**
     * The shape input type. Accessible only at compile time for type inference.
     */
    readonly input: I;
    /**
     * The shape output type. Accessible only at compile time for type inference.
     */
    readonly output: O;
    /**
     * `true` if the shape allows only {@linkcode parseAsync} and throws an error if {@linkcode parse} is called.
     * `false` if the shape can be used in both sync and async contexts.
     */
    readonly async: boolean;
    /**
     * Synchronously parses the value and returns {@linkcode Ok} or {@linkcode Err} object that wraps the result.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The {@linkcode Ok} instance if parsing has succeeded or {@linkcode Err} if parsing has failed.
     * @throws `Error` if the shape doesn't support the sync parsing, see {@linkcode async}.
     */
    try(input: unknown, options?: ParseOptions): Ok<O> | Err;
    /**
     * Asynchronously parses the value and returns {@linkcode Ok} or {@linkcode Err} object that wraps the result.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The {@linkcode Ok} instance if parsing has succeeded or {@linkcode Err} if parsing has failed.
     */
    tryAsync(input: unknown, options?: ParseOptions): Promise<Ok<O> | Err>;
    /**
     * Synchronously parses the value.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The value that conforms the output type of the shape.
     * @throws `Error` if the shape doesn't support the sync parsing, see {@linkcode async}.
     * @throws {@linkcode ValidationError} if any issues occur during parsing.
     */
    parse(input: unknown, options?: ParseOptions): O;
    /**
     * Asynchronously parses the value.
     *
     * @param input The value to parse.
     * @param options Parsing options.
     * @returns The value that conforms the output type of the shape.
     * @throws {@linkcode ValidationError} if any issues occur during parsing.
     */
    parseAsync(input: unknown, options?: ParseOptions): Promise<O>;
    /**
     * Synchronously parses the value and returns `undefined` if parsing fails.
     *
     * @param input The value to parse.
     * @returns The value that conforms the output type of the shape.
     * @throws `Error` if the shape doesn't support the sync parsing, see {@linkcode async}.
     */
    parseOrDefault(input: unknown): O | undefined;
    /**
     * Synchronously parses the value and returns the default value if parsing fails.
     *
     * @param input The value to parse.
     * @param defaultValue The default value that is returned if parsing fails.
     * @param options Parsing options.
     * @returns The value that conforms the output type of the shape.
     * @throws `Error` if the shape doesn't support the sync parsing, see {@linkcode async}.
     */
    parseOrDefault<T>(input: unknown, defaultValue: T, options?: ParseOptions): O | T;
    /**
     * Asynchronously parses the value and returns `undefined` value if parsing fails.
     *
     * @param input The value to parse.
     * @returns The value that conforms the output type of the shape.
     */
    parseOrDefaultAsync(input: unknown): Promise<O | undefined>;
    /**
     * Asynchronously parses the value and returns the default value if parsing fails.
     *
     * @param input The value to parse.
     * @param defaultValue The default value that is returned if parsing fails.
     * @param options Parsing options.
     * @returns The value that conforms the output type of the shape.
     */
    parseOrDefaultAsync<T>(input: unknown, defaultValue: T, options?: ParseOptions): Promise<O | T>;
}
/**
 * The shape that applies a transformer callback to the base shape output.
 *
 * @template S The base shape.
 * @template O The transformed value.
 */
declare class TransformShape<S extends AnyShape, O> extends Shape<S['input'], O> {
    /**
     * The base shape which output value is transformed.
     */
    readonly shape: S;
    /**
     * The callback that transforms the shape output value.
     *
     * @param output The {@linkcode shape} output value that must be transformed.
     * @param options Parsing options.
     * @return The transformation result.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    readonly callback: (output: S['output'], options: Readonly<ParseOptions>) => Promise<O> | O;
    /**
     * Creates the new {@linkcode TransformShape} instance.
     *
     * @param shape The base shape.
     * @param async If `true` then the transformed shape would wait for the promise returned from the callback to be
     * fulfilled. Otherwise, the value that is synchronously returned from the callback is used as an output.
     * @param callback The callback that transforms the shape output value.
     * @template S The base shape.
     * @template O The transformed value.
     */
    constructor(
    /**
     * The base shape which output value is transformed.
     */
    shape: S, async: boolean, 
    /**
     * The callback that transforms the shape output value.
     *
     * @param output The {@linkcode shape} output value that must be transformed.
     * @param options Parsing options.
     * @return The transformation result.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    callback: (output: S['output'], options: Readonly<ParseOptions>) => Promise<O> | O);
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _getInputValues(): unknown[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<O>;
    protected _applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<O>>;
}
/**
 * The shape that parses the output of the input shape with the output shape.
 *
 * @template I The input shape.
 * @template O The output shape.
 */
declare class PipeShape<I extends AnyShape, O extends Shape<I['output'], any>> extends Shape<I['input'], O['output']> {
    /**
     * The input shape.
     */
    readonly inputShape: I;
    /**
     * The output shape.
     */
    readonly outputShape: O;
    /**
     * Creates the new {@linkcode PipeShape} instance.
     *
     * @param inputShape The shape that parses the input value.
     * @param outputShape The shape that parses the output of `inputShape`.
     * @template I The input shape.
     * @template O The output shape.
     */
    constructor(
    /**
     * The input shape.
     */
    inputShape: I, 
    /**
     * The output shape.
     */
    outputShape: O);
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _getInputValues(): unknown[];
    protected _apply(input: unknown, options: ParseOptions): ApplyResult<O['output']>;
    protected _applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<O['output']>>;
}
/**
 * The shape that replaces an input value with an output value.
 *
 * @template S The shape that parses the input without the replaced value.
 * @template A The input value to replace.
 * @template B The output value.
 */
declare class ReplaceShape<S extends AnyShape, A, B> extends Shape<S['input'] | A, Exclude<S['output'], A> | B> {
    /**
     * The shape that parses the input without the replaced value.
     */
    readonly shape: S;
    /**
     * The input value to replace.
     */
    readonly inputValue: A;
    /**
     * The output value that is returned if an {@linkcode inputValue} is received.
     */
    readonly outputValue: B;
    private _result;
    /**
     * Creates the new {@linkcode ReplaceShape} instance.
     *
     * @param shape The shape that parses the input without the replaced value.
     * @param inputValue The input value to replace.
     * @param outputValue The output value that is returned if an `inputValue` is received.
     * @template S The shape that parses the input without the replaced value.
     * @template A The input value to replace.
     * @template B The output value.
     */
    constructor(
    /**
     * The shape that parses the input without the replaced value.
     */
    shape: S, 
    /**
     * The input value to replace.
     */
    inputValue: A, 
    /**
     * The output value that is returned if an {@linkcode inputValue} is received.
     */
    outputValue: B);
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _getInputValues(): unknown[];
    protected _apply(input: unknown, options: ParseOptions): ApplyResult<Exclude<S['output'], A> | B>;
    protected _applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<Exclude<S['output'], A> | B>>;
}
/**
 * The shape that excludes a value from both input and output.
 *
 * @template S The base shape.
 * @template T The excluded value.
 */
declare class ExcludeShape<S extends AnyShape, T> extends Shape<Exclude<S['input'], T>, Exclude<S['output'], T>> {
    /**
     * The base shape.
     */
    readonly shape: S;
    /**
     * The excluded value.
     */
    readonly excludedValue: T;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates the new {@linkcode ExcludeShape} instance.
     *
     * @param shape The shape that parses the input without the replaced value.
     * @param excludedValue The excluded value.
     * @param options The constraint options or an issue message.
     * @template S The shape that parses the input without the replaced value.
     * @template T The excluded value.
     */
    constructor(
    /**
     * The base shape.
     */
    shape: S, 
    /**
     * The excluded value.
     */
    excludedValue: T, options?: TypeConstraintOptions | Message);
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _getInputValues(): unknown[];
    protected _apply(input: unknown, options: ParseOptions): ApplyResult<Exclude<S['output'], T>>;
    protected _applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<Exclude<S['output'], T>>>;
}
/**
 * The shape that returns the fallback value if parsing fails.
 *
 * @template S The shape that parses the input.
 */
declare class CatchShape<S extends AnyShape, T> extends Shape<S['input'], S['output'] | T> {
    /**
     * The base shape.
     */
    readonly shape: S;
    /**
     * The value or a callback that returns a value that is returned if parsing has failed.
     */
    readonly fallback: T | (() => T);
    private _resultProvider;
    /**
     * Creates the new {@linkcode CatchShape} instance.
     *
     * @param shape The shape that parses the input.
     * @param fallback The value or a callback that returns a value that is returned if parsing has failed.
     * @template S The shape that parses the input.
     */
    constructor(
    /**
     * The base shape.
     */
    shape: S, 
    /**
     * The value or a callback that returns a value that is returned if parsing has failed.
     */
    fallback: T | (() => T));
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _getInputValues(): unknown[];
    protected _apply(input: unknown, options: ParseOptions): ApplyResult<S['output'] | T>;
    protected _applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<S['output'] | T>>;
}

/**
 * The shape which value can be coerced to a proper type during parsing.
 *
 * @template I The input value.
 * @template O The output value.
 */
declare class CoercibleShape<I = any, O = I> extends Shape<I, O> {
    protected _coerced: boolean;
    /**
     * Enables input value coercion.
     *
     * @returns The clone of the shape.
     */
    coerce(): this;
}

type InferTuple<U extends readonly AnyShape[], C extends 'input' | 'output'> = ToArray<{
    [K in keyof U]: U[K] extends AnyShape ? U[K][C] : never;
}>;
type InferArray<U extends readonly AnyShape[] | null, R extends AnyShape | null, C extends 'input' | 'output'> = U extends readonly AnyShape[] ? R extends AnyShape ? [...InferTuple<U, C>, ...R[C][]] : InferTuple<U, C> : R extends AnyShape ? R[C][] : any[];
type ToArray<T> = T extends any[] ? T : never;
/**
 * The shape of an array or a tuple.
 *
 * @template U The list of positioned element shapes, or `null` if there are no positioned elements.
 * @template R The shape of rest elements, or `null` if there are no rest elements.
 */
declare class ArrayShape<U extends readonly AnyShape[] | null, R extends AnyShape | null> extends CoercibleShape<InferArray<U, R, 'input'>, InferArray<U, R, 'output'>> {
    /**
     * The list of positioned element shapes or `null` if there are no positioned elements.
     */
    readonly shapes: U;
    /**
     * The shape of rest elements or `null` if there are no rest elements.
     */
    readonly restShape: R;
    protected _options: Message | TypeConstraintOptions | undefined;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode ArrayShape} instance.
     *
     * @param shapes The list of positioned element shapes or `null` if there are no positioned elements.
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @param options The type constraint options or the type issue message.
     * @template U The list of positioned element shapes, or `null` if there are no positioned elements.
     * @template R The shape of rest elements, or `null` if there are no rest elements.
     */
    constructor(
    /**
     * The list of positioned element shapes or `null` if there are no positioned elements.
     */
    shapes: U, 
    /**
     * The shape of rest elements or `null` if there are no rest elements.
     */
    restShape: R, options?: TypeConstraintOptions | Message);
    at(key: unknown): AnyShape | null;
    /**
     * Returns an array shape that has rest elements constrained by the given shape.
     *
     * The returned object shape would have no checks.
     *
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @returns The new array shape.
     * @template T The shape of rest elements.
     */
    rest<T extends AnyShape | null>(restShape: T): ArrayShape<U, T>;
    /**
     * Constrains the array length.
     *
     * @param length The minimum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    length(length: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the minimum array length.
     *
     * @param length The minimum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    min(length: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the maximum array length.
     *
     * @param length The maximum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    max(length: number, options?: ConstraintOptions | Message): this;
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<InferArray<U, R, 'output'>>;
    protected _applyAsync(input: any, options: ParseOptions): Promise<ApplyResult<InferArray<U, R, 'output'>>>;
    /**
     * Coerces value to an array or returns `null` if coercion isn't possible.
     *
     * @param value The non-array value to coerce.
     */
    protected _coerce(value: unknown): unknown[] | null;
}

/**
 * The shape of the bigint value.
 */
declare class BigIntShape extends CoercibleShape<bigint> {
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode BigIntShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    constructor(options?: TypeConstraintOptions | Message);
    protected _getInputTypes(): ValueType[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<bigint>;
    protected _coerce(value: any): bigint | null;
}

/**
 * The shape of the boolean value.
 */
declare class BooleanShape extends CoercibleShape<boolean> {
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode BooleanShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    constructor(options?: TypeConstraintOptions | Message);
    protected _getInputTypes(): ValueType[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<boolean>;
    /**
     * Coerces value to a boolean or returns `null` if coercion isn't possible.
     *
     * @param value The non-boolean value to coerce.
     */
    protected _coerce(value: unknown): boolean | null;
}

/**
 * The shape that constrains an input to be the exact value.
 *
 * @template T The value.
 */
declare class ConstShape<T> extends Shape<T> {
    readonly value: T;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    protected _typePredicate: (input: unknown) => boolean;
    /**
     * Creates a new {@linkcode ConstShape} instance.
     *
     * @param value The exact value.
     * @param options The type constraint options or an issue message.
     * @template T Allowed values.
     */
    constructor(value: T, options?: TypeConstraintOptions | Message);
    protected _getInputTypes(): ValueType[];
    protected _getInputValues(): unknown[];
    protected _apply(input: unknown, options: ParseOptions): ApplyResult<T>;
}

/**
 * The shape of the `Date` object.
 */
declare class DateShape extends CoercibleShape<Date> {
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode DateShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    constructor(options?: TypeConstraintOptions | Message);
    protected _getInputTypes(): ValueType[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<Date>;
    /**
     * Coerces value to a `Date` or returns `null` if coercion isn't possible.
     *
     * @param value The non-`Date` value to coerce.
     */
    protected _coerce(value: unknown): Date | null;
}

/**
 * The shape that constrains an input to one of values.
 *
 * @template T Allowed values.
 */
declare class EnumShape<T> extends CoercibleShape<T> {
    /**
     * The list of allowed values, a const key-value mapping, or an enum object.
     */
    readonly source: readonly T[] | ReadonlyDict<T>;
    /**
     * The list of unique values allowed as an input.
     */
    readonly values: readonly T[];
    /**
     * Key-value mapping passes as a source to constructor, or `null` if the source was a list of values.
     */
    protected _valueMapping: ReadonlyDict<T> | null;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode EnumShape} instance.
     *
     * @param source The list of allowed values, a const key-value mapping, or an enum object.
     * @param options The type constraint options or an issue message.
     * @template T Allowed values.
     */
    constructor(
    /**
     * The list of allowed values, a const key-value mapping, or an enum object.
     */
    source: readonly T[] | ReadonlyDict<T>, options?: TypeConstraintOptions | Message);
    protected _getInputTypes(): ValueType[];
    protected _getInputValues(): unknown[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<T>;
    /**
     * Coerces value to an enum value or returns {@linkcode UNRECOGNIZED} if coercion isn't possible.
     *
     * @param value The non-enum value to coerce.
     */
    protected _coerce(value: any): unknown;
}

/**
 * The shape of the class instance.
 *
 * @template C The class constructor.
 */
declare class InstanceShape<C extends new (...args: any[]) => any> extends Shape<InstanceType<C>> {
    readonly ctor: C;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode InstanceShape} instance.
     *
     * @param ctor The class constructor.
     * @param options The type constraint options or the type issue message.
     * @template C The class constructor.
     */
    constructor(ctor: C, options?: TypeConstraintOptions | Message);
    protected _getInputTypes(): ValueType[];
    protected _apply(input: unknown, options: ParseOptions): ApplyResult<InstanceType<C>>;
}

type ToIntersection<U> = (U extends any ? (k: U) => void : never) extends (k: infer I) => void ? I : never;
declare class IntersectionShape<U extends readonly AnyShape[]> extends Shape<ToIntersection<U[number]['input']>, ToIntersection<U[number]['output']>> {
    readonly shapes: U;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    constructor(shapes: U, options?: TypeConstraintOptions | Message);
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _apply(input: unknown, options: ParseOptions): ApplyResult<ToIntersection<U[number]['output']>>;
    protected _applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<ToIntersection<U[number]['output']>>>;
}

/**
 * The shape of a value deserialized from a JSON string.
 */
declare class JSONShape extends Shape<string, any> {
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    protected _jsonIssueFactory: (input: unknown, options: Readonly<ParseOptions>, param: unknown) => Issue[];
    protected _reviver?: (this: any, key: string, value: any) => any;
    /**
     * Creates a new {@linkcode JSONShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    constructor(options?: TypeConstraintOptions | Message);
    /**
     * Prescribes how each value originally produced by parsing is transformed before being returned.
     *
     * @param reviver The reviver callback.
     * @returns The clone of the shape.
     */
    revive(
    /**
     * Prescribes how each value originally produced by parsing is transformed before being returned.
     *
     * @param key The key associated with the value.
     * @param value The value produced by parsing.
     * @returns The transformed value.
     */
    reviver: (this: any, key: string, value: any) => any): this;
    protected _getInputTypes(): ValueType[];
    protected _apply(input: unknown, options: ParseOptions): ApplyResult;
}

/**
 * Lazily resolves a shape using the provider.
 *
 * @template S The base shape.
 */
declare class LazyShape<S extends AnyShape> extends Shape<S['input'], S['output']> {
    protected _shapeProvider: () => S;
    /**
     * Creates a new {@linkcode LazyShape} instance.
     *
     * @param shapeProvider The provider that returns the base shape.
     * @template S The base shape.
     */
    constructor(shapeProvider: () => S);
    /**
     * The base shape.
     */
    get shape(): S;
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _getInputValues(): unknown[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<S['output']>;
    protected _applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<S['output']>>;
}

/**
 * The shape of a `Map` instance.
 *
 * @template K The key shape.
 * @template V The value shape.
 */
declare class MapShape<K extends AnyShape, V extends AnyShape> extends CoercibleShape<Map<K['input'], V['input']>, Map<K['output'], V['output']>> {
    /**
     * The key shape.
     */
    readonly keyShape: K;
    /**
     * The value shape.
     */
    readonly valueShape: V;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode MapShape} instance.
     *
     * @param keyShape The key shape.
     * @param valueShape The value shape.
     * @param options The type constraint options or an issue message.
     * @template K The key shape.
     * @template V The value shape.
     */
    constructor(
    /**
     * The key shape.
     */
    keyShape: K, 
    /**
     * The value shape.
     */
    valueShape: V, options?: TypeConstraintOptions | Message);
    at(key: unknown): AnyShape | null;
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<Map<K['output'], V['output']>>;
    protected _applyAsync(input: any, options: ParseOptions): Promise<ApplyResult<Map<K['output'], V['output']>>>;
    /**
     * Coerces value to a list of `Map` entries, or returns `null` if coercion isn't possible.
     *
     * @param value The non-`Map` value to coerce.
     */
    protected _coerceEntries(value: any): [unknown, unknown][] | null;
}

/**
 * The shape that doesn't match any input.
 */
declare class NeverShape extends Shape<never> {
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode NeverShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    constructor(options?: TypeConstraintOptions | Message);
    protected _getInputTypes(): ValueType[];
    protected _apply(input: unknown, options: ParseOptions): ApplyResult<never>;
}

declare function isObjectLike(value: unknown): boolean;
declare function isNumber(value: unknown): boolean;
/**
 * Returns a function that creates a new array with a single issue.
 *
 * @param code The code of the issue.
 * @param defaultMessage The default message that is used if message isn't provided through options.
 * @param options Options provided by the user.
 * @param param The param that is added to the issue.
 * @returns The callback that takes an input and returns an array with a single issue.
 */
declare function createIssueFactory(code: unknown, defaultMessage: MessageCallback | string, options: TypeConstraintOptions | Message | undefined, param: unknown): (input: unknown, options: Readonly<ParseOptions>) => Issue[];
/**
 * Returns a function that creates a new array with a single issue.
 *
 * @param code The code of the issue.
 * @param defaultMessage The default message that is used if message isn't provided through options.
 * @param options Options provided by the user.
 * @returns The callback that takes an input and a param, and returns an array with a single issue.
 */
declare function createIssueFactory(code: unknown, defaultMessage: Message, options: TypeConstraintOptions | Message | undefined): (input: unknown, options: Readonly<ParseOptions>, param: unknown) => Issue[];

/**
 * The shape that constrains the input as a number.
 */
declare class NumberShape extends CoercibleShape<number> {
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    protected _typePredicate: typeof isNumber;
    /**
     * Creates a new {@linkcode NumberShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    constructor(options?: TypeConstraintOptions | Message);
    /**
     * Constrains the number to be greater than zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    positive(options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be less than zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    negative(options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be less or equal to zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    nonPositive(options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be greater or equal to zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    nonNegative(options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be greater than the value.
     *
     * @param value The exclusive minimum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    gt(value: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be less than the value.
     *
     * @param value The exclusive maximum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    lt(value: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be greater than or equal to the value.
     *
     * @param value The inclusive minimum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    gte(value: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be less than or equal to the value.
     *
     * @param value The inclusive maximum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    lte(value: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be a multiple of the divisor.
     *
     * @param value The positive number by which the input should be divisible without a remainder.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    multipleOf(value: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be a finite number.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    finite(options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be an integer.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    integer(options?: ConstraintOptions | Message): this;
    /**
     * Allows `NaN` as an input and output value, or replaces an input `NaN` value with a default output value.
     *
     * @param [defaultValue = NaN] The value that is used instead of `NaN` in the output.
     */
    nan(defaultValue?: number): Shape<number>;
    protected _getInputTypes(): ValueType[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<number>;
    /**
     * Coerces value to a number (not `NaN`) or returns `null` if coercion isn't possible.
     *
     * @param value The non-number value to coerce.
     */
    protected _coerce(value: any): number | null;
}
interface NumberShape {
    /**
     * Constrains the number to be greater than or equal to the value.
     *
     * Alias for {@linkcode gte}.
     *
     * @param value The inclusive minimum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    min(value: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the number to be less than or equal to the value.
     *
     * Alias for {@linkcode lte}.
     *
     * @param value The inclusive maximum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    max(value: number, options?: ConstraintOptions | Message): this;
}

type InferObject<P extends ReadonlyDict<AnyShape>, R extends AnyShape | null, C extends 'input' | 'output'> = Squash<UndefinedAsOptional<{
    [K in keyof P]: P[K][C];
}> & InferIndexer<R, C>>;
type InferIndexer<R extends AnyShape | null, C extends 'input' | 'output'> = R extends Shape ? {
    [key: string]: R[C];
} : unknown;
type StringKeyof<T extends object> = Extract<keyof T, string>;
type Squash<T> = T extends never ? never : {
    [K in keyof T]: T[K];
};
type UndefinedAsOptional<T> = OmitBy<T, undefined> & Partial<PickBy<T, undefined>>;
type OmitBy<T, V> = Omit<T, {
    [K in keyof T]: V extends Extract<T[K], V> ? K : never;
}[keyof T]>;
type PickBy<T, V> = Pick<T, {
    [K in keyof T]: V extends Extract<T[K], V> ? K : never;
}[keyof T]>;
type Optional<P extends ReadonlyDict<AnyShape>> = {
    [K in keyof P]: OpaqueReplaceShape<P[K], undefined>;
};
type Required<P extends ReadonlyDict<AnyShape>> = {
    [K in keyof P]: OpaqueExcludeShape<P[K], undefined>;
};
type KeysMode = 'preserved' | 'stripped' | 'exact';
/**
 * The shape of an object.
 *
 * @template P The mapping from an object key to a corresponding value shape.
 * @template R The shape that constrains values of
 * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
 */
declare class ObjectShape<P extends ReadonlyDict<AnyShape>, R extends AnyShape | null> extends Shape<InferObject<P, R, 'input'>, InferObject<P, R, 'output'>> {
    /**
     * The mapping from an object key to a corresponding value shape.
     */
    readonly shapes: P;
    /**
     * The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    readonly restShape: R;
    /**
     * The array of known object keys.
     */
    readonly keys: readonly StringKeyof<P>[];
    /**
     * The mode of unknown keys handling.
     */
    readonly keysMode: KeysMode;
    protected _options: Message | TypeConstraintOptions | undefined;
    protected _valueShapes: Shape[];
    protected _typePredicate: typeof isObjectLike;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    protected _exactIssueFactory?: (input: unknown, options: Readonly<ParseOptions>, param: unknown) => Issue[];
    /**
     * Creates a new {@linkcode ObjectShape} instance.
     *
     * @param shapes The mapping from an object key to a corresponding value shape.
     * @param restShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures). If `null`
     * then values thea fall under the index signature are unconstrained.
     * @param options The type constraint options or an issue message.
     * @param keysMode The mode of unknown keys handling.
     * @template P The mapping from an object key to a corresponding value shape.
     * @template R The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    constructor(
    /**
     * The mapping from an object key to a corresponding value shape.
     */
    shapes: P, 
    /**
     * The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    restShape: R, options?: TypeConstraintOptions | Message, keysMode?: KeysMode);
    at(key: any): AnyShape | null;
    /**
     * Merge properties from the other object shape.
     *
     * If a property with the same key already exists on this object shape then it is overwritten. The index signature of
     * this shape and its {@linkcode keysMode} is preserved intact.
     *
     * The returned object shape would have no checks.
     *
     * @param shape The object shape which properties must be added to this object shape.
     * @returns The new object shape.
     * @template T Properties to add.
     */
    extend<T extends ReadonlyDict<AnyShape>>(shape: ObjectShape<T, any>): ObjectShape<Pick<P, Exclude<keyof P, keyof T>> & T, R>;
    /**
     * Add properties to an object shape.
     *
     * If a property with the same key already exists on this object shape then it is overwritten. The index signature of
     * this shape and its {@linkcode keysMode} is preserved intact.
     *
     * The returned object shape would have no checks.
     *
     * @param shapes The properties to add.
     * @returns The new object shape.
     * @template T The shapes of properties to add.
     */
    extend<T extends ReadonlyDict<AnyShape>>(shapes: T): ObjectShape<Pick<P, Exclude<keyof P, keyof T>> & T, R>;
    /**
     * Returns an object shape that only has properties with listed keys.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to pick.
     * @returns The new object shape.
     * @template K The tuple of keys to pick.
     */
    pick<K extends StringKeyof<P>[]>(keys: K): ObjectShape<Pick<P, K[number]>, R>;
    /**
     * Returns an object shape that doesn't have the listed keys.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to omit.
     * @returns The new object shape.
     * @template K The tuple of keys to omit.
     */
    omit<K extends StringKeyof<P>[]>(keys: K): ObjectShape<Omit<P, K[number]>, R>;
    /**
     * Returns an object shape with all properties marked as optional.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    partial(): ObjectShape<Optional<P>, R>;
    /**
     * Returns an object shape with keys marked as optional.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to make optional.
     * @returns The new object shape.
     * @template K The list of string keys.
     */
    partial<K extends StringKeyof<P>[]>(keys: K): ObjectShape<Omit<P, K[number]> & Optional<Pick<P, K[number]>>, R>;
    /**
     * Returns an object shape with all properties marked as required.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    required(): ObjectShape<Required<P>, R>;
    /**
     * Returns an object shape with keys marked as required.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to make required.
     * @returns The new object shape.
     * @template K The list of string keys.
     */
    required<K extends StringKeyof<P>[]>(keys: K): ObjectShape<Omit<P, K[number]> & Required<Pick<P, K[number]>>, R>;
    /**
     * Returns an object shape that allows only known keys and has no index signature.
     *
     * The returned object shape would have no checks.
     *
     * @param options The constraint options or an issue message.
     * @returns The new object shape.
     */
    exact(options?: TypeConstraintOptions | Message): ObjectShape<P, null>;
    /**
     * Returns an object shape that doesn't have an index signature and all unknown keys are stripped.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    strip(): ObjectShape<P, null>;
    /**
     * Returns an object shape that has an index signature that doesn't constrain values.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    preserve(): ObjectShape<P, null>;
    /**
     * Returns an object shape that has an index signature that is constrained by the given shape.
     *
     * The returned object shape would have no checks.
     *
     * @param restShape The shape that validates values at unknown keys.
     * @returns The new object shape.
     * @template T The index signature shape.
     */
    rest<T extends AnyShape>(restShape: T): ObjectShape<P, T>;
    /**
     * Returns the enum shape of keys of this object.
     */
    keyof(): EnumShape<StringKeyof<P>>;
    /**
     * Constrains an object to be an `Object` instance or to have a `null` prototype.
     */
    plain(): this;
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<InferObject<P, R, 'output'>>;
    protected _applyAsync(input: any, options: ParseOptions): Promise<ApplyResult<InferObject<P, R, 'output'>>>;
    /**
     * Unknown keys are preserved as is and aren't checked.
     */
    private _applyRestUnchecked;
    /**
     * Unknown keys are either parsed with a {@linkcode restShape}, stripped, or cause an issue.
     */
    private _applyRestChecked;
}

/**
 * The shape of a value wrapped in a `Promise` instance.
 *
 * @template S The shape of the resolved value.
 */
declare class PromiseShape<S extends AnyShape> extends CoercibleShape<Promise<S['input']>, Promise<S['output']>> {
    readonly shape: S;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode PromiseShape} instance.
     *
     * @param shape The shape of the resolved value.
     * @param options The type constraint options or the type issue message.
     * @template S The shape of the resolved value.
     */
    constructor(shape: S, options?: TypeConstraintOptions | Message);
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _apply(input: unknown, options: ParseOptions): ApplyResult<Promise<S['output']>>;
    protected _applyAsync(input: any, options: ParseOptions): Promise<ApplyResult<Promise<S['output']>>>;
}

type InferRecord<K extends Shape<string, PropertyKey> | null, V extends AnyShape, C extends 'input' | 'output'> = undefined extends V[C] ? Partial<Record<K extends Shape ? K[C] : string, V[C]>> : Record<K extends Shape ? K[C] : string, V[C]>;
/**
 * The shape that describes an object with string keys and values that conform the given shape.
 *
 * @template K The key shape.
 * @template V The value shape.
 */
declare class RecordShape<K extends Shape<string, PropertyKey> | null, V extends AnyShape> extends Shape<InferRecord<K, V, 'input'>, InferRecord<K, V, 'output'>> {
    /**
     * The key shape or `null` if keys are preserved intact.
     */
    readonly keyShape: K;
    /**
     * The value shape.
     */
    readonly valueShape: V;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode RecordShape} instance.
     *
     * @param keyShape The key shape, or `null` if keys should be preserved intact.
     * @param valueShape The value shape.
     * @param options The type constraint options or an issue message.
     * @template K The key shape.
     * @template V The value shape.
     */
    constructor(
    /**
     * The key shape or `null` if keys are preserved intact.
     */
    keyShape: K, 
    /**
     * The value shape.
     */
    valueShape: V, options?: TypeConstraintOptions | Message);
    at(key: unknown): AnyShape | null;
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<InferRecord<K, V, 'output'>>;
    protected _applyAsync(input: any, options: ParseOptions): Promise<ApplyResult<InferRecord<K, V, 'output'>>>;
}

/**
 * The shape of a `Set` instance.
 *
 * @template S The value shape.
 */
declare class SetShape<S extends AnyShape> extends CoercibleShape<Set<S['input']>, Set<S['output']>> {
    /**
     * The value shape
     */
    readonly shape: S;
    protected _options: Message | TypeConstraintOptions | undefined;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode SetShape} instance.
     *
     * @param shape The value shape
     * @param options The type constraint options or the type issue message.
     * @template S The value shape.
     */
    constructor(
    /**
     * The value shape
     */
    shape: S, options?: TypeConstraintOptions | Message);
    at(key: unknown): AnyShape | null;
    /**
     * Constrains the set size.
     *
     * @param size The minimum set size.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    size(size: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the minimum set size.
     *
     * @param size The minimum set size.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    min(size: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the maximum set size.
     *
     * @param size The maximum set size.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    max(size: number, options?: ConstraintOptions | Message): this;
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<Set<S['output']>>;
    protected _applyAsync(input: any, options: ParseOptions): Promise<ApplyResult<Set<S['output']>>>;
    /**
     * Coerces value to a list of `Set` values, or returns `null` if coercion isn't possible.
     *
     * @param value The non-`Set` value to coerce.
     */
    protected _coerceValues(value: unknown): unknown[] | null;
}

/**
 * The shape that constrains the input as a string.
 */
declare class StringShape extends CoercibleShape<string> {
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode StringShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    constructor(options?: TypeConstraintOptions | Message);
    /**
     * The shortcut to apply both {@linkcode min} and {@linkcode max} constraints.
     *
     * @param length The exact length a string must have.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    length(length: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the string length to be greater than or equal to the length.
     *
     * @param length The minimum string length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    min(length: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the string length to be less than or equal to the length.
     *
     * @param length The maximum string length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    max(length: number, options?: ConstraintOptions | Message): this;
    /**
     * Constrains the string to match a regexp.
     *
     * @param re The regular expression that the sting must conform.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    regex(re: RegExp, options?: ConstraintOptions | Message): this;
    protected _getInputTypes(): ValueType[];
    protected _apply(input: any, options: ParseOptions): ApplyResult<string>;
    /**
     * Coerces value to a string or returns `null` if coercion isn't possible.
     *
     * @param value The non-string value to coerce.
     */
    protected _coerce(value: any): string | null;
}

/**
 * The shape of the arbitrary symbol.
 */
declare class SymbolShape extends Shape<symbol> {
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>) => Issue[];
    /**
     * Creates a new {@linkcode SymbolShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    constructor(options?: TypeConstraintOptions | Message);
    protected _getInputTypes(): ValueType[];
    protected _apply(input: unknown, options: ParseOptions): ApplyResult<symbol>;
}

/**
 * Returns the list of shapes that are applicable to the input.
 */
type LookupCallback = (input: any) => readonly AnyShape[];
/**
 * The shape that requires an input to conform at least one of shapes.
 *
 * @template U The list of shapes that comprise a union.
 */
declare class UnionShape<U extends readonly AnyShape[]> extends Shape<U[number]['input'], U[number]['output']> {
    /**
     * The list of shapes that comprise a union.
     */
    readonly shapes: U;
    protected _options: Message | TypeConstraintOptions | undefined;
    protected _typeIssueFactory: (input: unknown, options: Readonly<ParseOptions>, param: unknown) => Issue[];
    /**
     * Creates a new {@linkcode UnionShape} instance.
     *
     * @param shapes The list of shapes that comprise a union.
     * @param options The union constraint options or an issue message.
     * @template U The list of shapes that comprise a union.
     */
    constructor(
    /**
     * The list of shapes that comprise a union.
     */
    shapes: U, options?: TypeConstraintOptions | Message);
    protected get _lookup(): LookupCallback;
    at(key: unknown): AnyShape | null;
    protected _requiresAsync(): boolean;
    protected _getInputTypes(): ValueType[];
    protected _getInputValues(): unknown[];
    protected _apply(input: unknown, options: ParseOptions): ApplyResult<U[number]['output']>;
    protected _applyAsync(input: unknown, options: ParseOptions): Promise<ApplyResult<U[number]['output']>>;
}

/**
 * Creates the unconstrained shape.
 *
 * You can specify compile-time type to enhance type inference.
 *
 * @template T The input and output.
 */
declare function any<T = any>(): Shape<T>;
/**
 * Creates a shape that is constrained with a [narrowing predicate](https://www.typescriptlang.org/docs/handbook/2/narrowing.html).
 *
 * @param cb The type predicate that returns `true` if value conforms the required type, or `false` otherwise.
 * @param options The constraint options or an issue message.
 * @returns The shape that has the narrowed output.
 * @template I The input value.
 * @template O The output value.
 */
declare function any<I = any, O extends I = I>(cb: (value: I) => value is O, options?: ConstraintOptions | Message): Shape<I, O>;
/**
 * Creates a shape that is constrained with a predicate.
 *
 * @param cb The predicate that returns truthy result if value is valid, or returns falsy result otherwise.
 * @param options The constraint options or an issue message.
 * @template I The input value.
 * @template O The output value.
 */
declare function any<I = any, O = I>(cb: (value: I) => boolean, options?: ConstraintOptions | Message): Shape<I, O>;

/**
 * Creates the unconstrained array shape.
 *
 * @param options The constraint options or an issue message.
 */
declare function array(options?: TypeConstraintOptions | Message): ArrayShape<null, null>;
/**
 * Creates the array shape with elements that conform the element shape.
 *
 * @param shape The shape of an array element.
 * @param options The constraint options or an issue message.
 * @template S The shape of array elements.
 */
declare function array<S extends AnyShape | null>(shape: S, options?: TypeConstraintOptions | Message): ArrayShape<null, S>;

/**
 * Creates the bigint shape.
 *
 * @param options The constraint options or an issue message.
 */
declare function bigint(options?: TypeConstraintOptions | Message): BigIntShape;

/**
 * Creates the boolean shape.
 *
 * @param options The constraint options or an issue message.
 */
declare function boolean(options?: TypeConstraintOptions | Message): BooleanShape;

/**
 * Creates the constant value shape.
 *
 * @param value The value to which the input must be strictly equal.
 * @param options The constraint options or an issue message.
 * @template T The value type.
 */
declare function const_<T extends Literal>(value: T, options?: TypeConstraintOptions | Message): ConstShape<T>;

/**
 * Creates the `Date` shape.
 *
 * @param options The constraint options or an issue message.
 */
declare function date(options?: TypeConstraintOptions | Message): DateShape;

/**
 * Creates the shape that constrains input with the list of values.
 *
 * @param values The list of values allowed for the input.
 * @param options The constraint options or an issue message.
 * @template T Allowed values.
 * @template U The list of allowed values.
 */
declare function enum_<T extends Literal, U extends readonly [T, ...T[]]>(values: U, options?: TypeConstraintOptions | Message): EnumShape<U[number]>;
/**
 * Creates the shape that constrains input with values of [the enum-like object](https://www.typescriptlang.org/docs/handbook/enums.html).
 *
 * @param valueMapping The native enum or a mapping object.
 * @param options The constraint options or an issue message.
 * @template T Allowed values.
 * @template U The object that maps from the key to an enum value.
 */
declare function enum_<T extends Literal, U extends ReadonlyDict<T>>(valueMapping: U, options?: TypeConstraintOptions | Message): EnumShape<U[keyof U]>;

/**
 * Creates the finite number shape.
 *
 * @param options The constraint options or an issue message.
 */
declare function finite(options?: TypeConstraintOptions | Message): NumberShape;

/**
 * Creates the class instance shape.
 *
 * @param ctor The instance constructor.
 * @param options The constraint options or an issue message.
 * @template C The instance constructor.
 */
declare function instanceOf<C extends new (...args: any[]) => any>(ctor: C, options?: TypeConstraintOptions | Message): InstanceShape<C>;

/**
 * Creates an intersection shape that tries to parse the input with all provided shapes.
 *
 * @param shapes The list of shapes.
 * @param options The constraint options or an issue message.
 * @template U The tuple of intersected shapes.
 */
declare function intersection<U extends [AnyShape, ...AnyShape[]]>(shapes: U, options?: TypeConstraintOptions | Message): IntersectionShape<U>;

/**
 * Creates the shape that parses JSON strings.
 *
 * @param options The constraint options or an issue message.
 */
declare function json(options?: TypeConstraintOptions | Message): JSONShape;

/**
 * Creates the shape that resolves the underlying shape on-demand.
 *
 * @param shapeProvider The provider that returns the base shape.
 */
declare function lazy<S extends AnyShape>(shapeProvider: () => S): LazyShape<S>;

/**
 * Creates the `Map` instance shape.
 *
 * @param keyShape The key shape.
 * @param valueShape The value shape.
 * @param options The type constraint options or an issue message.
 * @template K The key shape.
 * @template V The value shape.
 */
declare function map<K extends AnyShape, V extends AnyShape>(keyShape: K, valueShape: V, options?: TypeConstraintOptions | Message): MapShape<K, V>;

/**
 * Creates the shape that requires an input to equal to `NaN`.
 *
 * @param options The constraint options or an issue message.
 */
declare function nan(options?: TypeConstraintOptions | Message): ConstShape<number>;

/**
 * Creates the integer shape.
 *
 * @param options The constraint options or an issue message.
 */
declare function integer(options?: TypeConstraintOptions | Message): NumberShape;

/**
 * Creates the shape that always raises an issue.
 *
 * @param options The constraint options or an issue message.
 */
declare function never(options?: TypeConstraintOptions | Message): NeverShape;

/**
 * Creates the shape that requires an input to equal to `null`.
 *
 * @param options The constraint options or an issue message.
 */
declare function null_(options?: TypeConstraintOptions | Message): ConstShape<null>;

/**
 * Creates the number shape.
 *
 * @param options The constraint options or an issue message.
 */
declare function number(options?: TypeConstraintOptions | Message): NumberShape;

/**
 * Creates the object shape.
 *
 * @param shapes The mapping from an object key to a corresponding shape.
 * @param options The constraint options or an issue message.
 */
declare function object<P extends ReadonlyDict<AnyShape>>(shapes: P, options?: TypeConstraintOptions | Message): ObjectShape<P, null>;

/**
 * Creates the `Promise` instance shape.
 *
 * @param shape The shape of the resolved value.
 * @param options The constraint options or an issue message.
 * @template S The shape of the resolved value.
 */
declare function promise<S extends AnyShape>(shape: S, options?: TypeConstraintOptions | Message): PromiseShape<S>;

/**
 * Creates the shape that describes an object with string keys and values that conform the given shape.
 *
 * @param valueShape The shape of the record values.
 * @param options The constraint options or an issue message.
 * @template V The value shape.
 */
declare function record<V extends AnyShape>(valueShape: V, options?: TypeConstraintOptions | Message): RecordShape<null, V>;
/**
 * Creates the shape that describes an object with string keys and values that conform the given shape.
 *
 * @param keyShape The shape of record keys.
 * @param valueShape The shape of the record values.
 * @param options The constraint options or an issue message.
 * @template K The key shape.
 * @template V The value shape.
 */
declare function record<K extends Shape<string, PropertyKey>, V extends AnyShape>(keyShape: K, valueShape: V, options?: TypeConstraintOptions | Message): RecordShape<K, V>;

/**
 * Creates the `Set` instance shape.
 *
 * @param shape The value shape
 * @param options The constraint options or an issue message.
 * @template S The value shape.
 */
declare function set<S extends AnyShape>(shape: S, options?: TypeConstraintOptions | Message): SetShape<S>;

/**
 * Creates the string shape.
 *
 * @param options The constraint options or an issue message.
 */
declare function string(options?: TypeConstraintOptions | Message): StringShape;

/**
 * Creates the symbol shape.
 *
 * @param options The constraint options or an issue message.
 */
declare function symbol(options?: TypeConstraintOptions | Message): SymbolShape;

/**
 * Creates the shape that synchronously transforms the input value.
 *
 * @param cb The callback that processes the input value.
 * @template I The input value.
 * @template O The output value.
 */
declare function transform<I = any, O = any>(cb: (input: I, options: Readonly<ParseOptions>) => O): Shape<I, O>;
/**
 * Creates the shape that asynchronously transforms the input value.
 *
 * @param cb The callback that processes the input value.
 * @template I The input value.
 * @template O The output value.
 */
declare function transformAsync<I = any, O = any>(cb: (input: I, options: Readonly<ParseOptions>) => Promise<O>): Shape<I, O>;

/**
 * Creates the tuple shape.
 *
 * @param shapes The list of tuple element shapes.
 * @param options The constraint options or an issue message.
 * @template U The tuple elements.
 */
declare function tuple<U extends readonly [AnyShape, ...AnyShape[]]>(shapes: U, options?: TypeConstraintOptions | Message): ArrayShape<U, null>;
/**
 * Creates the tuple shape with rest elements.
 *
 * @param shapes The list of tuple element shapes.
 * @param restShape The shape of rest elements.
 * @param options The constraint options or an issue message.
 * @template U The head tuple elements.
 * @template R The rest tuple elements.
 */
declare function tuple<U extends readonly [AnyShape, ...AnyShape[]], R extends AnyShape | null>(shapes: U, restShape: R, options?: TypeConstraintOptions | Message): ArrayShape<U, R>;

/**
 * Creates the shape that requires an input to equal to `undefined`.
 *
 * @param options The constraint options or an issue message.
 */
declare function undefined_(options?: TypeConstraintOptions | Message): ConstShape<undefined>;

/**
 * Creates a union shape that tries to parse the input with one of the provided shapes.
 *
 * @param shapes The list of shapes to try.
 * @param options The constraint options or an issue message.
 * @template U The tuple of united shapes.
 */
declare function union<U extends [AnyShape, ...AnyShape[]]>(shapes: U, options?: TypeConstraintOptions | Message): UnionShape<U>;

/**
 * Creates the unconstrained shape with unknown value.
 */
declare function unknown(): Shape<unknown>;

/**
 * Creates a shape that requires an input to be `undefined` at runtime and typed as `void`.
 *
 * @param options The constraint options or an issue message.
 */
declare function void_(options?: TypeConstraintOptions | Message): ConstShape<void>;

/**
 * Guards each argument of a function with a corresponding shape.
 *
 * ```
 * const callbackFactory = d.guard([d.string(), d.boolean()]);
 *
 * callbackFactory((arg1: string, arg2: boolean) => …);
 * ```
 *
 * @param argShapes The array of shapes that parse arguments.
 * @template U The arguments array.
 * @returns The factory that guards functions.
 */
declare function guard<U extends readonly [AnyShape, ...AnyShape[]] | []>(argShapes: U): <R, T = any>(callback: (this: T, ...args: InferTuple<U, 'output'>) => R) => (this: T, ...args: InferTuple<U, 'input'>) => R;
/**
 * Guards a function arguments with an array/tuple shape.
 *
 * ```
 * const callbackFactory = d.guard(d.array(d.string()));
 *
 * callbackFactory((...args: string[]) => …);
 * ```
 *
 * @param argsShape The shape that parses an arguments array.
 * @template I The array of input arguments.
 * @template O The array of output arguments.
 * @returns The factory that guards functions.
 */
declare function guard<I extends readonly any[], O extends readonly any[]>(argsShape: Shape<I, O>): <R, T = any>(callback: (this: T, ...args: O) => R) => (this: T, ...args: I) => R;
/**
 * Guards an argument of an arity 1 function with a shape.
 *
 * ```
 * const callbackFactory = d.guard(d.string());
 *
 * callbackFactory((arg: string) => …);
 * ```
 *
 * @param argShape The shape that parses a single argument.
 * @template I The input argument.
 * @template O The output argument.
 * @returns The factory that guards functions.
 */
declare function guard<I, O>(argShape: Shape<I, O>): <R, T = any>(callback: (this: T, arg: O) => R) => (this: T, arg: I) => R;
/**
 * Guards each argument of a function with a corresponding shape.
 *
 * ```
 * const callback = d.guard(
 *   [d.string(), d.boolean()],
 *   (arg1: string, arg2: boolean) => …
 * );
 * ```
 *
 * @param argShapes The array of shapes that parse arguments.
 * @param callback The callback to guard.
 * @template U The arguments array.
 * @template R The returned value.
 * @template T The value of `this`.
 * @returns The guarded callback.
 */
declare function guard<U extends readonly [AnyShape, ...AnyShape[]] | [], R, T = any>(argShapes: U, callback: (this: T, ...args: InferTuple<U, 'output'>) => R): (this: T, ...args: InferTuple<U, 'input'>) => R;
/**
 * Guards a function arguments with an array/tuple shape.
 *
 * ```
 * const callback = d.guard(
 *   d.array(d.string()),
 *   (...args: string[]) => …
 * );
 * ```
 *
 * @param argsShape The shape that parses an arguments array.
 * @param callback The callback to guard.
 * @template I The array of input arguments.
 * @template O The array of output arguments.
 * @template R The returned value.
 * @template T The value of `this`.
 * @returns The guarded callback.
 */
declare function guard<I extends readonly any[], O extends readonly any[], R, T = any>(argsShape: Shape<I, O>, callback: (this: T, ...args: O) => R): (this: T, ...args: I) => R;
/**
 * Guards an argument of an arity 1 function with a shape.
 *
 * ```
 * const callback = d.guard(
 *   d.string(),
 *   (arg: string) => …
 * );
 * ```
 *
 * @param argShape The shape that parses a single argument.
 * @param callback The callback to guard.
 * @template I The input argument.
 * @template O The output argument.
 * @template R The returned value.
 * @template T The value of `this`.
 * @returns The guarded callback.
 */
declare function guard<I, O, R, T = any>(argShape: Shape<I, O>, callback: (this: T, arg: O) => R): (this: T, arg: I) => R;
/**
 * Guards each argument of a function with a corresponding shape.
 *
 * ```
 * const callbackFactory = d.guardAsync([d.string(), d.boolean()]);
 *
 * callbackFactory((arg1: string, arg2: boolean) => …);
 * ```
 *
 * @param argShapes The array of shapes that parse arguments.
 * @template U The arguments array.
 * @returns The factory that guards functions.
 */
declare function guardAsync<U extends readonly [AnyShape, ...AnyShape[]] | []>(argShapes: U): <R, T = any>(callback: (this: T, ...args: InferTuple<U, 'output'>) => Promise<R> | R) => (this: T, ...args: InferTuple<U, 'input'>) => Promise<R>;
/**
 * Guards a function arguments with an array/tuple shape.
 *
 * ```
 * const callbackFactory = d.guardAsync(d.array(d.string()));
 *
 * callbackFactory((...args: string[]) => …);
 * ```
 *
 * @param argsShape The shape that parses an arguments array.
 * @template I The array of input arguments.
 * @template O The array of output arguments.
 * @returns The factory that guards functions.
 */
declare function guardAsync<I extends readonly any[], O extends readonly any[]>(argsShape: Shape<I, O>): <R, T = any>(callback: (this: T, ...args: O) => Promise<R> | R) => (this: T, ...args: I) => Promise<R>;
/**
 * Guards an argument of an arity 1 function with a shape.
 *
 * ```
 * const callbackFactory = d.guardAsync(d.string());
 *
 * callbackFactory((arg: string) => …);
 * ```
 *
 * @param argShape The shape that parses a single argument.
 * @template I The input argument.
 * @template O The output argument.
 * @returns The factory that guards functions.
 */
declare function guardAsync<I, O>(argShape: Shape<I, O>): <R, T = any>(callback: (this: T, arg: O) => Promise<R> | R) => (this: T, arg: I) => Promise<R>;
/**
 * Guards each argument of a function with a corresponding shape.
 *
 * ```
 * const callback = d.guardAsync(
 *   [d.string(), d.boolean()],
 *   (arg1: string, arg2: boolean) => …
 * );
 * ```
 *
 * @param argShapes The array of shapes that parse arguments.
 * @param callback The callback to guard.
 * @template U The arguments array.
 * @template R The returned value.
 * @template T The value of `this`.
 * @returns The guarded callback.
 */
declare function guardAsync<U extends readonly [AnyShape, ...AnyShape[]] | [], R, T = any>(argShapes: U, callback: (this: T, ...args: InferTuple<U, 'output'>) => Promise<R> | R): (this: T, ...args: InferTuple<U, 'input'>) => Promise<R>;
/**
 * Guards a function arguments with an array/tuple shape.
 *
 * ```
 * const callback = d.guardAsync(
 *   d.array(d.string()),
 *   (...args: string[]) => …
 * );
 * ```
 *
 * @param argsShape The shape that parses an arguments array.
 * @param callback The callback to guard.
 * @template I The array of input arguments.
 * @template O The array of output arguments.
 * @template R The returned value.
 * @template T The value of `this`.
 * @returns The guarded callback.
 */
declare function guardAsync<I extends readonly any[], O extends readonly any[], R, T = any>(argsShape: Shape<I, O>, callback: (this: T, ...args: O) => Promise<R> | R): (this: T, ...args: I) => Promise<R>;
/**
 * Guards an argument of an arity 1 function with a shape.
 *
 * ```
 * const callback = d.guardAsync(
 *   d.string(),
 *   (arg: string) => …
 * );
 * ```
 *
 * @param argShape The shape that parses a single argument.
 * @param callback The callback to guard.
 * @template I The input argument.
 * @template O The output argument.
 * @template R The returned value.
 * @template T The value of `this`.
 * @returns The guarded callback.
 */
declare function guardAsync<I, O, R, T = any>(argShape: Shape<I, O>, callback: (this: T, arg: O) => Promise<R> | R): (this: T, arg: I) => Promise<R>;

/**
 * An error thrown if parsing failed. Custom checkers and transformers can throw this error to notify that the operation
 * has failed.
 */
declare class ValidationError extends Error {
    /**
     * The list of issues that caused the error.
     */
    issues: Issue[];
    /**
     * Creates a new {@linkcode ValidationError} instance.
     *
     * @param issues The mutable list of partially defined issues that have caused an error.
     */
    constructor(issues: Partial<Issue>[]);
}

export { AnyShape, ArrayShape, BigIntShape, BooleanShape, CatchShape, Check, CheckCallback, CheckOptions, CoercibleShape, ConstShape, ConstraintOptions, DateShape, EnumShape, Err, ExcludeShape, InstanceShape, IntersectionShape, Issue, JSONShape, KeysMode, LazyShape, Literal, MapShape, Message, MessageCallback, NeverShape, NumberShape, ObjectShape, Ok, OpaqueExcludeShape, OpaqueReplaceShape, ParseOptions, PipeShape, PromiseShape, RecordShape, ReplaceShape, SetShape, Shape, StringShape, SymbolShape, TransformShape, TypeConstraintOptions, UnionShape, ValidationError, ValueType, intersection as and, any, array, bigint, boolean as bool, boolean, const_ as const, createIssueFactory, date, enum_ as enum, finite, guard, guardAsync, instanceOf, integer as int, integer, intersection, json, lazy, map, nan, never, null_ as null, number, object, union as or, promise, record, set, string, symbol, transform, transformAsync, tuple, undefined_ as undefined, union, unknown, void_ as void };
