function __extends(constructor, superConstructor) {
    var Super = function () {
        this.constructor = constructor;
    };
    Super.prototype = superConstructor.prototype;
    constructor.prototype = new Super();
}

/**
 * An error thrown if parsing failed. Custom checkers and transformers can throw this error to notify that the operation
 * has failed.
 */
var ValidationError = /** @class */ (function (_super) {
    __extends(ValidationError, _super);
    /**
     * Creates a new {@linkcode ValidationError} instance.
     *
     * @param issues The mutable list of partially defined issues that have caused an error.
     */
    function ValidationError(issues) {
        var _newTarget = this.constructor;
        var _this = this;
        var message = '';
        for (var _i = 0, issues_1 = issues; _i < issues_1.length; _i++) {
            var issue = issues_1[_i];
            inflateIssue(issue);
            message +=
                '\n' + issue.code + ' at /' + issue.path.join('/') + (issue.message != null ? ': ' + issue.message : '');
        }
        _this = _super.call(this, message) || this;
        Object.setPrototypeOf(_this, _newTarget.prototype);
        _this.name = 'ValidationError';
        _this.issues = issues;
        return _this;
    }
    return ValidationError;
}(Error));
function inflateIssue(issue) {
    if (issue.code == null) {
        issue.code = 'unknown';
    }
    if (issue.path == null) {
        issue.path = [];
    }
}

var ERROR_REQUIRES_ASYNC = 'Shape is async, consider using tryAsync, parseAsync, or parseOrDefaultAsync';
var ERROR_SHAPE_EXPECTED = 'Provider must return a shape';
var ERROR_FORBIDDEN_AT_RUNTIME = 'Must not be accessed at runtime';
var ERROR_ASYNC_ARGUMENTS = 'Some of argument shapes are async, consider using guardAsync';
var CODE_ARRAY_MIN = 'arrayMinLength';
var CODE_ARRAY_MAX = 'arrayMaxLength';
var CODE_CONST = 'const';
var CODE_ENUM = 'enum';
var CODE_EXCLUSION = 'exclusion';
var CODE_INSTANCE = 'instance';
var CODE_INTERSECTION = 'intersection';
var CODE_JSON = 'json';
var CODE_PREDICATE = 'predicate';
var CODE_NUMBER_INTEGER = 'numberInteger';
var CODE_NUMBER_FINITE = 'numberFinite';
var CODE_NUMBER_GT = 'numberGreaterThan';
var CODE_NUMBER_GTE = 'numberGreaterThanOrEqual';
var CODE_NUMBER_LT = 'numberLessThan';
var CODE_NUMBER_LTE = 'numberLessThanOrEqual';
var CODE_NUMBER_MULTIPLE_OF = 'numberMultipleOf';
var CODE_SET_MIN = 'setMinSize';
var CODE_SET_MAX = 'setMaxSize';
var CODE_STRING_MIN = 'stringMinLength';
var CODE_STRING_MAX = 'stringMaxLength';
var CODE_STRING_REGEX = 'stringRegex';
var CODE_TYPE = 'type';
var CODE_TUPLE = 'tuple';
var CODE_UNION = 'union';
var CODE_UNKNOWN_KEYS = 'unknownKeys';
var TYPE_OBJECT = 'object';
var TYPE_ARRAY = 'array';
var TYPE_STRING = 'string';
var TYPE_SYMBOL = 'symbol';
var TYPE_NUMBER = 'number';
var TYPE_BIGINT = 'bigint';
var TYPE_BOOLEAN = 'boolean';
var TYPE_NULL = 'null';
var TYPE_UNDEFINED = 'undefined';
var TYPE_ANY = 'any';
var TYPE_NEVER = 'never';
var TYPE_PROMISE = 'promise';
var TYPE_SET = 'set';
var TYPE_MAP = 'map';
var TYPE_DATE = 'date';
var MESSAGE_ARRAY_TYPE = 'Must be an array';
var MESSAGE_ARRAY_MIN = 'Must have the minimum length of %s';
var MESSAGE_ARRAY_MAX = 'Must have the maximum length of %s';
var MESSAGE_SET_MIN = 'Must have the minimum size of %s';
var MESSAGE_SET_MAX = 'Must have the maximum size of %s';
var MESSAGE_BIGINT_TYPE = 'Must be a bigint';
var MESSAGE_BOOLEAN_TYPE = 'Must be a boolean';
var MESSAGE_CONST = 'Must be equal to %s';
var MESSAGE_DATE_TYPE = 'Must be a Date';
var MESSAGE_ENUM = 'Must be equal to one of %s';
var MESSAGE_EXCLUSION = 'Must not be equal to %s';
var MESSAGE_INSTANCE = 'Must be a class instance';
var MESSAGE_INTERSECTION = 'Must conform the intersection';
var MESSAGE_PREDICATE = 'Must conform the predicate';
var MESSAGE_MAP_TYPE = 'Must be a Map';
var MESSAGE_NEVER_TYPE = 'Must not be used';
var MESSAGE_NUMBER_TYPE = 'Must be a number';
var MESSAGE_NUMBER_INTEGER = 'Must be an integer';
var MESSAGE_NUMBER_FINITE = 'Must be an finite number';
var MESSAGE_NUMBER_GT = 'Must be greater than %s';
var MESSAGE_NUMBER_GTE = 'Must be greater than or equal to %s';
var MESSAGE_NUMBER_LT = 'Must be less than %s';
var MESSAGE_NUMBER_LTE = 'Must be less than or equal to %s';
var MESSAGE_NUMBER_MULTIPLE_OF = 'Must be a multiple of %s';
var MESSAGE_OBJECT_TYPE = 'Must be an object';
var MESSAGE_PROMISE_TYPE = 'Must be a Promise';
var MESSAGE_SET_TYPE = 'Must be a Set';
var MESSAGE_UNKNOWN_KEYS = 'Must not have unknown keys: %s';
var MESSAGE_STRING_TYPE = 'Must be a string';
var MESSAGE_STRING_MIN = 'Must have the minimum length of %s';
var MESSAGE_STRING_MAX = 'Must have the maximum length of %s';
var MESSAGE_STRING_REGEX = 'Must match the pattern %s';
var MESSAGE_SYMBOL_TYPE = 'Must be a symbol';
var MESSAGE_TUPLE = 'Must be a tuple of length %s';
var MESSAGE_UNION = 'Must conform the union of %s';

function getValueType(value) {
    var type = typeof value;
    if (type !== 'object') {
        return type;
    }
    if (value === null) {
        return TYPE_NULL;
    }
    if (isArray(value)) {
        return TYPE_ARRAY;
    }
    if (value instanceof Date) {
        return TYPE_DATE;
    }
    return type;
}
function ok(value) {
    return { ok: true, value: value };
}
var isArray = Array.isArray;
var getPrototypeOf = Object.getPrototypeOf;
function isEqual(a, b) {
    return a === b || (a !== a && b !== b);
}
function isObjectLike(value) {
    return value !== null && typeof value === 'object';
}
function isPlainObject(value) {
    var prototype;
    return isObjectLike(value) && ((prototype = getPrototypeOf(value)) === null || prototype.constructor === Object);
}
function isIterable(value) {
    return isObjectLike(value) && (typeof value[Symbol.iterator] === 'function' || typeof value.length === 'number');
}
function isNumber(value) {
    return typeof value === 'number' && value === value;
}
function isDate(value) {
    return value instanceof Date && (value = value.getTime()) === value;
}
function isAsyncShapes(shapes) {
    var async = false;
    for (var i = 0; i < shapes.length && !async; ++i) {
        async = shapes[i].async;
    }
    return async;
}
function isUnsafeCheck(check) {
    return check.unsafe;
}
/**
 * Returns an array index, or -1 if key isn't an index.
 */
function toArrayIndex(key) {
    if (typeof key === 'string' && '' + +key === key) {
        key = +key;
    }
    return Number.isInteger(key) && key >= 0 && key < 0xffffffff ? key : -1;
}
/**
 * The convenient shortcut to add built-in checks to shapes.
 */
function setCheck(shape, key, options, param, cb) {
    return shape.check(cb, {
        key: key,
        unsafe: options !== null && typeof options === 'object' && options.unsafe,
        param: param,
    });
}
function createIssueFactory(code, defaultMessage, options, param) {
    var paramKnown = arguments.length === 4;
    var meta;
    var message = defaultMessage;
    if (options !== null && typeof options === 'object') {
        if (options.message !== undefined) {
            message = options.message;
        }
        meta = options.meta;
    }
    else if (typeof options === 'function') {
        message = options;
    }
    else if (options != null) {
        message = options;
    }
    if (typeof message === 'function') {
        if (paramKnown) {
            return function (input, options) { return [
                { code: code, path: [], input: input, message: message(param, code, input, meta, options), param: param, meta: meta },
            ]; };
        }
        else {
            return function (input, options, param) { return [
                { code: code, path: [], input: input, message: message(param, code, input, meta, options), param: param, meta: meta },
            ]; };
        }
    }
    if (typeof message === 'string') {
        if (paramKnown) {
            message = message.replace('%s', param);
        }
        else if (message.indexOf('%s') !== -1) {
            return function (input, options, param) { return [{ code: code, path: [], input: input, message: message.replace('%s', param), param: param, meta: meta }]; };
        }
    }
    if (paramKnown) {
        return function (input) { return [{ code: code, path: [], input: input, message: message, param: param, meta: meta }]; };
    }
    else {
        return function (input, options, param) { return [{ code: code, path: [], input: input, message: message, param: param, meta: meta }]; };
    }
}
function unshiftPath(issues, key) {
    var issuesLength = issues.length;
    for (var i = 0; i < issuesLength; ++i) {
        issues[i].path.unshift(key);
    }
}
function concatIssues(issues, result) {
    if (issues !== null) {
        issues.push.apply(issues, result);
        return issues;
    }
    return result;
}
function captureIssues(error) {
    if (error instanceof ValidationError) {
        return error.issues;
    }
    throw error;
}
function enableBitAt(bits, index) {
    if (typeof bits === 'number') {
        if (index < 32) {
            return bits | (1 << index);
        }
        bits = [bits, 0, 0];
    }
    bits[index >> 5] |= 1 << index % 32;
    return bits;
}
function isBitEnabledAt(bits, index) {
    if (typeof bits === 'number') {
        return bits >>> index !== 0;
    }
    else {
        return bits[index >> 5] >>> index % 32 !== 0;
    }
}
function setKeyValue(obj, key, value) {
    if (key === '__proto__') {
        Object.defineProperty(obj, key, { value: value, writable: true, enumerable: true, configurable: true });
    }
    else {
        obj[key] = value;
    }
}
/**
 * Returns the shallow clone of the instance object.
 */
function cloneObject(instance) {
    return Object.assign(Object.create(getPrototypeOf(instance)), instance);
}
/**
 * Clones the first `keyCount` keys of a plain object.
 */
function cloneObjectEnumerableKeys(input, keyCount) {
    if (keyCount === void 0) { keyCount = -1; }
    var output = {};
    if (keyCount < 0) {
        for (var key in input) {
            setKeyValue(output, key, input[key]);
        }
    }
    if (keyCount > 0) {
        var index = 0;
        for (var key in input) {
            if (index === keyCount) {
                break;
            }
            setKeyValue(output, key, input[key]);
            ++index;
        }
    }
    return output;
}
/**
 * Clones known keys of the object.
 */
function cloneObjectKnownKeys(input, keys) {
    var output = {};
    var keysLength = keys.length;
    for (var i = 0; i < keysLength; ++i) {
        var key = keys[i];
        if (key in input) {
            setKeyValue(output, key, input[key]);
        }
    }
    return output;
}
function createApplyChecksCallback(checks) {
    var checksLength = checks.length;
    if (checksLength === 0) {
        return null;
    }
    if (checksLength === 1) {
        var _a = checks[0], unsafe0_1 = _a.unsafe, cb0_1 = _a.callback;
        return function (output, issues, options) {
            if (issues === null || unsafe0_1) {
                var result = void 0;
                try {
                    result = cb0_1(output, options);
                }
                catch (error) {
                    return concatIssues(issues, captureIssues(error));
                }
                if (result != null) {
                    return appendIssue(issues, result);
                }
            }
            return issues;
        };
    }
    if (checksLength === 2) {
        var _b = checks[0], unsafe0_2 = _b.unsafe, cb0_2 = _b.callback, _c = checks[1], unsafe1_1 = _c.unsafe, cb1_1 = _c.callback;
        return function (output, issues, options) {
            if (issues === null || unsafe0_2) {
                var result = void 0;
                try {
                    result = cb0_2(output, options);
                }
                catch (error) {
                    issues = concatIssues(issues, captureIssues(error));
                    if (!options.verbose) {
                        return issues;
                    }
                }
                if (result != null) {
                    issues = appendIssue(issues, result);
                    if (issues !== null && !options.verbose) {
                        return issues;
                    }
                }
            }
            if (issues === null || unsafe1_1) {
                var result = void 0;
                try {
                    result = cb1_1(output, options);
                }
                catch (error) {
                    issues = concatIssues(issues, captureIssues(error));
                }
                if (result != null) {
                    issues = appendIssue(issues, result);
                }
            }
            return issues;
        };
    }
    return function (output, issues, options) {
        for (var i = 0; i < checksLength; ++i) {
            var _a = checks[i], unsafe = _a.unsafe, callback = _a.callback;
            var result = void 0;
            if (issues !== null && !unsafe) {
                continue;
            }
            try {
                result = callback(output, options);
            }
            catch (error) {
                issues = concatIssues(issues, captureIssues(error));
                if (!options.verbose) {
                    return issues;
                }
            }
            if (result != null) {
                issues = appendIssue(issues, result);
                if (issues !== null && !options.verbose) {
                    return issues;
                }
            }
        }
        return issues;
    };
}
function appendIssue(issues, result) {
    if (isArray(result)) {
        var resultLength = result.length;
        if (resultLength === 0) {
            return issues;
        }
        for (var i = 0; i < resultLength; ++i) {
            inflateIssue(result[i]);
        }
        if (issues === null) {
            issues = result;
        }
        else {
            issues.push.apply(issues, result);
        }
    }
    else {
        inflateIssue(result);
        if (issues === null) {
            issues = [result];
        }
        else {
            issues.push(result);
        }
    }
    return issues;
}
function unique(arr) {
    var uniqueArr = null;
    for (var i = 0; i < arr.length; ++i) {
        var value = arr[i];
        if (arr.includes(value, i + 1)) {
            if (uniqueArr === null) {
                uniqueArr = arr.slice(0, i);
            }
            continue;
        }
        if (uniqueArr !== null) {
            uniqueArr.push(value);
        }
    }
    return uniqueArr || arr;
}
function returnFalse() {
    return false;
}
function returnTrue() {
    return true;
}
function returnArray() {
    return [];
}

var defaultParseOptions = Object.freeze({ verbose: false, coerced: false });
/**
 * The baseline shape implementation.
 *
 * @template I The input value.
 * @template O The output value.
 */
var Shape = /** @class */ (function () {
    function Shape() {
        /**
         * The human-readable shape description.
         */
        this.description = '';
        /**
         * A callback that applies checks to the given value.
         */
        this._applyChecks = null;
        /**
         * `true` if some checks from {@linkcode _checkMap} were marked as unsafe, `false` otherwise.
         */
        this._unsafe = false;
    }
    /**
     * Returns a sub-shape that describes a value associated with the given property name, or `null` if there's no such
     * sub-shape.
     *
     * @param key The key for which the sub-shape must be retrieved.
     * @returns The sub-shape or `null` if there's no such key in the shape.
     */
    Shape.prototype.at = function (key) {
        return null;
    };
    /**
     * Adds a human-readable description text to the shape.
     *
     * @param text The description text.
     * @returns The clone of the shape with the description added.
     */
    Shape.prototype.describe = function (text) {
        var shape = cloneObject(this);
        shape.description = text;
        return shape;
    };
    /**
     * Adds the check that is applied to the shape output.
     *
     * If the {@linkcode CheckOptions.key} is defined and there's already a check with the same key then it is replaced.
     * If the key is `undefined` then the `cb` identity is used as a key.
     *
     * If check callback returns an empty array, it is considered that no issues have occurred.
     *
     * @param cb The callback that checks the shape output.
     * @param options The check options.
     * @returns The clone of the shape with the check added.
     */
    Shape.prototype.check = function (cb, options) {
        if (options === void 0) { options = {}; }
        var _a = options.key, key = _a === void 0 ? cb : _a, _b = options.unsafe, unsafe = _b === void 0 ? false : _b, param = options.param;
        return this._replaceChecks(new Map(this._checkMap).set(key, { callback: cb, unsafe: unsafe, param: param }));
    };
    /**
     * Returns the {@linkcode Check} by its key.
     *
     * @param key The check key.
     * @returns The check or `undefined` if there's no check associated with the key.
     */
    Shape.prototype.getCheck = function (key) {
        var _a;
        return (_a = this._checkMap) === null || _a === void 0 ? void 0 : _a.get(key);
    };
    /**
     * Deletes the check from the shape.
     *
     * @param key The check key.
     * @returns The clone of the shape if check was deleted or this shape if there are no matching check.
     */
    Shape.prototype.deleteCheck = function (key) {
        var _checkMap = this._checkMap;
        if (_checkMap === undefined || !_checkMap.has(key)) {
            return this;
        }
        var checkMap = new Map(_checkMap);
        checkMap.delete(key);
        return this._replaceChecks(checkMap);
    };
    Shape.prototype.refine = function (cb, options) {
        var issueFactory = createIssueFactory(CODE_PREDICATE, MESSAGE_PREDICATE, options, cb);
        return setCheck(this, cb, options, cb, function (input, options) {
            if (!cb(input)) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Pipes the shape output to another shape.
     *
     * @param shape The shape that validates the output if this shape.
     * @returns The {@linkcode PipeShape} instance.
     * @template T The output value.
     */
    Shape.prototype.to = function (shape) {
        return new PipeShape(this, shape);
    };
    /**
     * Synchronously transforms the output value of the shape with a transformer callback.
     *
     * @param cb The callback that transforms the shape output value.
     * @returns The {@linkcode TransformShape} instance.
     * @template T The output value.
     */
    Shape.prototype.transform = function (
    /**
     * @param output The shape output value.
     * @param options Parsing options.
     * @return The transformed value.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    cb) {
        return new TransformShape(this, false, cb);
    };
    /**
     * Asynchronously transforms the output value of the shape with a transformer callback.
     *
     * @param cb The callback that transforms the shape output value.
     * @returns The {@linkcode TransformShape} instance.
     * @template T The transformed value.
     */
    Shape.prototype.transformAsync = function (
    /**
     * @param output The shape output value.
     * @param options Parsing options.
     * @return The transformed value.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    cb) {
        return new TransformShape(this, true, cb);
    };
    /**
     * Adds a brand to the output type of the shape.
     *
     * @returns An opaque shape with branded output type.
     * @template T The unique brand marker. By default, this shape is used as a marker.
     */
    Shape.prototype.brand = function () {
        return this;
    };
    /**
     * Replaces an input value with an output value.
     *
     * @param inputValue The input value to replace.
     * @param outputValue The output value that is returned if an `inputValue` is received.
     * @returns The {@linkcode ReplaceShape} instance.
     * @template A The input value to replace.
     * @template B The output value.
     */
    Shape.prototype.replace = function (inputValue, outputValue) {
        return new ReplaceShape(this, inputValue, outputValue);
    };
    /**
     * Input value is passed directly to the output without any checks.
     *
     * @param value The included value.
     * @param options The constraint options or an issue message.
     * @returns The {@linkcode ReplaceShape} instance.
     * @template T The included value.
     */
    Shape.prototype.include = function (value, options) {
        return this.replace(value, value);
    };
    /**
     * Excludes value from both input and output.
     *
     * @param value The excluded value.
     * @param options The constraint options or an issue message.
     * @returns The {@linkcode ExcludeShape} instance.
     * @template T The excluded value.
     */
    Shape.prototype.exclude = function (value, options) {
        return new ExcludeShape(this, value, options);
    };
    Shape.prototype.optional = function (defaultValue) {
        return this.replace(undefined, defaultValue);
    };
    Shape.prototype.nullable = function (defaultValue) {
        return this.replace(null, arguments.length === 0 ? null : defaultValue);
    };
    Shape.prototype.nullish = function (defaultValue) {
        return this.nullable(arguments.length === 0 ? null : defaultValue).optional(defaultValue);
    };
    /**
     * Prevents an input and output from being `undefined`.
     *
     * @param options The constraint options or an issue message.
     * @returns The {@linkcode ExcludeShape} instance.
     */
    Shape.prototype.nonOptional = function (options) {
        return this.exclude(undefined, options);
    };
    Shape.prototype.catch = function (fallback) {
        return new CatchShape(this, fallback);
    };
    /**
     * Must return `true` if the shape must be used in async context only, otherwise the shape can be used in both sync
     * and async contexts. Override this method to implement a custom shape.
     */
    Shape.prototype._requiresAsync = function () {
        return false;
    };
    /**
     * Returns the list of runtime value types that can be processed by the shape. Used for various optimizations.
     */
    Shape.prototype._getInputTypes = function () {
        return [TYPE_ANY];
    };
    /**
     * Returns the list of all input values that are known beforehand.
     */
    Shape.prototype._getInputValues = function () {
        return [];
    };
    /**
     * Synchronously parses the input.
     *
     * Override this method to implement a custom shape.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @returns `null` if input matches the output, {@linkcode Ok} that wraps the output, or an array of captured issues.
     */
    Shape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    /**
     * Asynchronously parses the input.
     *
     * Override this method to implement a custom shape that requires an async execution context.
     *
     * @param input The shape input to parse.
     * @param options Parsing options.
     * @returns `null` if input matches the output, {@linkcode Ok} that wraps the output, or an array of captured issues.
     */
    Shape.prototype._applyAsync = function (input, options) {
        var _this = this;
        return new Promise(function (resolve) { return resolve(_this._apply(input, options)); });
    };
    /**
     * Returns a shape clone with new set of checks.
     *
     * @param checkMap The map from a check key to a corresponding check.
     * @returns The clone of the shape.
     */
    Shape.prototype._replaceChecks = function (checkMap) {
        var checks = Array.from(checkMap.values());
        var shape = cloneObject(this);
        shape._checkMap = checkMap;
        shape._applyChecks = createApplyChecksCallback(checks);
        shape._unsafe = checks.some(isUnsafeCheck);
        return shape;
    };
    return Shape;
}());
Object.defineProperties(Shape.prototype, {
    input: {
        get: function () {
            throw new Error(ERROR_FORBIDDEN_AT_RUNTIME);
        },
    },
    output: {
        get: function () {
            throw new Error(ERROR_FORBIDDEN_AT_RUNTIME);
        },
    },
    async: {
        get: function () {
            var async = this._requiresAsync();
            var _applyAsync = Shape.prototype['_applyAsync'];
            if (async) {
                this._apply = function () {
                    throw new Error(ERROR_REQUIRES_ASYNC);
                };
            }
            else if (this._applyAsync !== _applyAsync) {
                this._applyAsync = _applyAsync;
            }
            Object.defineProperty(this, 'async', { value: async });
            return async;
        },
    },
    try: {
        get: function () {
            var _this = this;
            void this.async;
            var cb = function (input, options) {
                var result = _this._apply(input, options || defaultParseOptions);
                if (result === null) {
                    return ok(input);
                }
                if (isArray(result)) {
                    return { ok: false, issues: result };
                }
                return ok(result.value);
            };
            Object.defineProperty(this, 'try', { value: cb });
            return cb;
        },
    },
    tryAsync: {
        get: function () {
            var _this = this;
            void this.async;
            var cb = function (input, options) {
                return _this._applyAsync(input, options || defaultParseOptions).then(function (result) {
                    if (result === null) {
                        return ok(input);
                    }
                    if (isArray(result)) {
                        return { ok: false, issues: result };
                    }
                    return result;
                });
            };
            Object.defineProperty(this, 'tryAsync', { value: cb });
            return cb;
        },
    },
    parse: {
        get: function () {
            var _this = this;
            void this.async;
            var cb = function (input, options) {
                var result = _this._apply(input, options || defaultParseOptions);
                if (result === null) {
                    return input;
                }
                if (isArray(result)) {
                    throw new ValidationError(result);
                }
                return result.value;
            };
            Object.defineProperty(this, 'parse', { value: cb });
            return cb;
        },
    },
    parseAsync: {
        get: function () {
            var _this = this;
            void this.async;
            var cb = function (input, options) {
                return _this._applyAsync(input, options || defaultParseOptions).then(function (result) {
                    if (result === null) {
                        return input;
                    }
                    if (isArray(result)) {
                        throw new ValidationError(result);
                    }
                    return result.value;
                });
            };
            Object.defineProperty(this, 'parseAsync', { value: cb });
            return cb;
        },
    },
    parseOrDefault: {
        get: function () {
            var _this = this;
            void this.async;
            var cb = function (input, defaultValue, options) {
                var result = _this._apply(input, options || defaultParseOptions);
                if (result === null) {
                    return input;
                }
                if (isArray(result)) {
                    return defaultValue;
                }
                return result.value;
            };
            Object.defineProperty(this, 'parseOrDefault', { value: cb });
            return cb;
        },
    },
    parseOrDefaultAsync: {
        get: function () {
            var _this = this;
            void this.async;
            var cb = function (input, defaultValue, options) {
                return _this._applyAsync(input, options || defaultParseOptions).then(function (result) {
                    if (result === null) {
                        return input;
                    }
                    if (isArray(result)) {
                        return defaultValue;
                    }
                    return result.value;
                });
            };
            Object.defineProperty(this, 'parseOrDefaultAsync', { value: cb });
            return cb;
        },
    },
});
/**
 * The shape that applies a transformer callback to the base shape output.
 *
 * @template S The base shape.
 * @template O The transformed value.
 */
var TransformShape = /** @class */ (function (_super) {
    __extends(TransformShape, _super);
    /**
     * Creates the new {@linkcode TransformShape} instance.
     *
     * @param shape The base shape.
     * @param async If `true` then the transformed shape would wait for the promise returned from the callback to be
     * fulfilled. Otherwise, the value that is synchronously returned from the callback is used as an output.
     * @param callback The callback that transforms the shape output value.
     * @template S The base shape.
     * @template O The transformed value.
     */
    function TransformShape(
    /**
     * The base shape which output value is transformed.
     */
    shape, async, 
    /**
     * The callback that transforms the shape output value.
     *
     * @param output The {@linkcode shape} output value that must be transformed.
     * @param options Parsing options.
     * @return The transformation result.
     * @throws {@linkcode ValidationError} to notify that the transformation cannot be successfully completed.
     */
    callback) {
        var _this = _super.call(this) || this;
        _this.shape = shape;
        _this.callback = callback;
        if (async) {
            _this._requiresAsync = returnTrue;
        }
        return _this;
    }
    TransformShape.prototype._requiresAsync = function () {
        return this.shape.async;
    };
    TransformShape.prototype._getInputTypes = function () {
        return this.shape['_getInputTypes']();
    };
    TransformShape.prototype._getInputValues = function () {
        return this.shape['_getInputValues']();
    };
    TransformShape.prototype._apply = function (input, options) {
        var _a = this, shape = _a.shape, callback = _a.callback, _applyChecks = _a._applyChecks;
        var issues = null;
        var output = input;
        var result = shape['_apply'](input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        try {
            output = callback(output, options);
        }
        catch (error) {
            return captureIssues(error);
        }
        if ((_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) && !isEqual(input, output)) {
            return ok(output);
        }
        return issues;
    };
    TransformShape.prototype._applyAsync = function (input, options) {
        var _a = this, shape = _a.shape, callback = _a.callback, _applyChecks = _a._applyChecks;
        return shape['_applyAsync'](input, options).then(function (result) {
            var output = input;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                output = result.value;
            }
            return new Promise(function (resolve) { return resolve(callback(output, options)); }).then(function (output) {
                var issues = null;
                if ((_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) &&
                    !isEqual(input, output)) {
                    return ok(output);
                }
                return issues;
            }, captureIssues);
        });
    };
    return TransformShape;
}(Shape));
/**
 * The shape that parses the output of the input shape with the output shape.
 *
 * @template I The input shape.
 * @template O The output shape.
 */
var PipeShape = /** @class */ (function (_super) {
    __extends(PipeShape, _super);
    /**
     * Creates the new {@linkcode PipeShape} instance.
     *
     * @param inputShape The shape that parses the input value.
     * @param outputShape The shape that parses the output of `inputShape`.
     * @template I The input shape.
     * @template O The output shape.
     */
    function PipeShape(
    /**
     * The input shape.
     */
    inputShape, 
    /**
     * The output shape.
     */
    outputShape) {
        var _this = _super.call(this) || this;
        _this.inputShape = inputShape;
        _this.outputShape = outputShape;
        return _this;
    }
    PipeShape.prototype._requiresAsync = function () {
        return this.inputShape.async || this.outputShape.async;
    };
    PipeShape.prototype._getInputTypes = function () {
        return this.inputShape['_getInputTypes']();
    };
    PipeShape.prototype._getInputValues = function () {
        return this.inputShape['_getInputValues']();
    };
    PipeShape.prototype._apply = function (input, options) {
        var _a = this, inputShape = _a.inputShape, outputShape = _a.outputShape, _applyChecks = _a._applyChecks;
        var issues;
        var output = input;
        var result = inputShape['_apply'](input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        var outputResult = outputShape['_apply'](output, options);
        if (outputResult !== null) {
            if (isArray(outputResult)) {
                return outputResult;
            }
            result = outputResult;
            output = outputResult.value;
        }
        if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
            return result;
        }
        return issues;
    };
    PipeShape.prototype._applyAsync = function (input, options) {
        var _a = this, inputShape = _a.inputShape, outputShape = _a.outputShape, _applyChecks = _a._applyChecks;
        var result = null;
        var output = input;
        return inputShape['_applyAsync'](input, options)
            .then(function (inputResult) {
            if (inputResult !== null) {
                if (isArray(inputResult)) {
                    return inputResult;
                }
                result = inputResult;
                output = inputResult.value;
            }
            return outputShape['_applyAsync'](output, options);
        })
            .then(function (outputResult) {
            var issues;
            if (outputResult !== null) {
                if (isArray(outputResult)) {
                    return outputResult;
                }
                result = outputResult;
                output = outputResult.value;
            }
            if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
                return result;
            }
            return issues;
        });
    };
    return PipeShape;
}(Shape));
/**
 * The shape that replaces an input value with an output value.
 *
 * @template S The shape that parses the input without the replaced value.
 * @template A The input value to replace.
 * @template B The output value.
 */
var ReplaceShape = /** @class */ (function (_super) {
    __extends(ReplaceShape, _super);
    /**
     * Creates the new {@linkcode ReplaceShape} instance.
     *
     * @param shape The shape that parses the input without the replaced value.
     * @param inputValue The input value to replace.
     * @param outputValue The output value that is returned if an `inputValue` is received.
     * @template S The shape that parses the input without the replaced value.
     * @template A The input value to replace.
     * @template B The output value.
     */
    function ReplaceShape(
    /**
     * The shape that parses the input without the replaced value.
     */
    shape, 
    /**
     * The input value to replace.
     */
    inputValue, 
    /**
     * The output value that is returned if an {@linkcode inputValue} is received.
     */
    outputValue) {
        var _this = _super.call(this) || this;
        _this.shape = shape;
        _this.inputValue = inputValue;
        _this.outputValue = outputValue;
        _this._result = isEqual(inputValue, outputValue) ? null : ok(outputValue);
        return _this;
    }
    ReplaceShape.prototype._requiresAsync = function () {
        return this.shape.async;
    };
    ReplaceShape.prototype._getInputTypes = function () {
        return this.shape['_getInputTypes']().concat(getValueType(this.inputValue));
    };
    ReplaceShape.prototype._getInputValues = function () {
        return this.shape['_getInputValues']().concat(this.inputValue);
    };
    ReplaceShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var issues;
        var output = input;
        var result = isEqual(input, this.inputValue) ? this._result : this.shape['_apply'](input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
            return result;
        }
        return issues;
    };
    ReplaceShape.prototype._applyAsync = function (input, options) {
        var _this = this;
        var _applyChecks = this._applyChecks;
        if (isEqual(input, this.inputValue)) {
            if (_applyChecks !== null) {
                return new Promise(function (resolve) {
                    var issues = _applyChecks(_this.outputValue, null, options);
                    resolve(issues !== null ? issues : _this._result);
                });
            }
            return Promise.resolve(this._result);
        }
        return this.shape['_applyAsync'](input, options).then(function (result) {
            var issues;
            var output = input;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                output = result.value;
            }
            if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
                return result;
            }
            return issues;
        });
    };
    return ReplaceShape;
}(Shape));
/**
 * The shape that excludes a value from both input and output.
 *
 * @template S The base shape.
 * @template T The excluded value.
 */
var ExcludeShape = /** @class */ (function (_super) {
    __extends(ExcludeShape, _super);
    /**
     * Creates the new {@linkcode ExcludeShape} instance.
     *
     * @param shape The shape that parses the input without the replaced value.
     * @param excludedValue The excluded value.
     * @param options The constraint options or an issue message.
     * @template S The shape that parses the input without the replaced value.
     * @template T The excluded value.
     */
    function ExcludeShape(
    /**
     * The base shape.
     */
    shape, 
    /**
     * The excluded value.
     */
    excludedValue, options) {
        var _this = _super.call(this) || this;
        _this.shape = shape;
        _this.excludedValue = excludedValue;
        _this._typeIssueFactory = createIssueFactory(CODE_EXCLUSION, MESSAGE_EXCLUSION, options, excludedValue);
        return _this;
    }
    ExcludeShape.prototype._requiresAsync = function () {
        return this.shape.async;
    };
    ExcludeShape.prototype._getInputTypes = function () {
        return this.shape['_getInputTypes']();
    };
    ExcludeShape.prototype._getInputValues = function () {
        var _this = this;
        return this.shape['_getInputValues']().filter(function (value) { return !isEqual(_this.excludedValue, value); });
    };
    ExcludeShape.prototype._apply = function (input, options) {
        var _a = this, excludedValue = _a.excludedValue, _typeIssueFactory = _a._typeIssueFactory, _applyChecks = _a._applyChecks;
        var issues;
        var output = input;
        if (isEqual(input, excludedValue)) {
            return _typeIssueFactory(input, options);
        }
        var result = this.shape['_apply'](input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
            if (isEqual(output, excludedValue)) {
                return _typeIssueFactory(input, options);
            }
        }
        if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
            return result;
        }
        return issues;
    };
    ExcludeShape.prototype._applyAsync = function (input, options) {
        var _a = this, excludedValue = _a.excludedValue, _typeIssueFactory = _a._typeIssueFactory, _applyChecks = _a._applyChecks;
        if (isEqual(input, excludedValue)) {
            return Promise.resolve(_typeIssueFactory(input, options));
        }
        return this.shape['_applyAsync'](input, options).then(function (result) {
            var issues;
            var output = input;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                output = result.value;
                if (isEqual(output, excludedValue)) {
                    return _typeIssueFactory(input, options);
                }
            }
            if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
                return result;
            }
            return issues;
        });
    };
    return ExcludeShape;
}(Shape));
/**
 * The shape that returns the fallback value if parsing fails.
 *
 * @template S The shape that parses the input.
 */
var CatchShape = /** @class */ (function (_super) {
    __extends(CatchShape, _super);
    /**
     * Creates the new {@linkcode CatchShape} instance.
     *
     * @param shape The shape that parses the input.
     * @param fallback The value or a callback that returns a value that is returned if parsing has failed.
     * @template S The shape that parses the input.
     */
    function CatchShape(
    /**
     * The base shape.
     */
    shape, 
    /**
     * The value or a callback that returns a value that is returned if parsing has failed.
     */
    fallback) {
        var _this = _super.call(this) || this;
        _this.shape = shape;
        _this.fallback = fallback;
        if (typeof fallback === 'function') {
            _this._resultProvider = function () { return ok(fallback()); };
        }
        else {
            var result_1 = ok(fallback);
            _this._resultProvider = function () { return result_1; };
        }
        return _this;
    }
    CatchShape.prototype._requiresAsync = function () {
        return this.shape.async;
    };
    CatchShape.prototype._getInputTypes = function () {
        return this.shape['_getInputTypes']();
    };
    CatchShape.prototype._getInputValues = function () {
        return this.shape['_getInputValues']();
    };
    CatchShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var result = this.shape['_apply'](input, options);
        var issues;
        var output = input;
        if (result !== null) {
            if (isArray(result)) {
                result = this._resultProvider();
            }
            output = result.value;
        }
        if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
            return result;
        }
        return issues;
    };
    CatchShape.prototype._applyAsync = function (input, options) {
        var _this = this;
        var _applyChecks = this._applyChecks;
        return this.shape['_applyAsync'](input, options).then(function (result) {
            var issues;
            var output = input;
            if (result !== null) {
                if (isArray(result)) {
                    result = _this._resultProvider();
                }
                output = result.value;
            }
            if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
                return result;
            }
            return issues;
        });
    };
    return CatchShape;
}(Shape));

/**
 * The shape which value can be coerced to a proper type during parsing.
 *
 * @template I The input value.
 * @template O The output value.
 */
var CoercibleShape = /** @class */ (function (_super) {
    __extends(CoercibleShape, _super);
    function CoercibleShape() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this._coerced = false;
        return _this;
    }
    /**
     * Enables input value coercion.
     *
     * @returns The clone of the shape.
     */
    CoercibleShape.prototype.coerce = function () {
        var shape = cloneObject(this);
        shape._coerced = true;
        return shape;
    };
    return CoercibleShape;
}(Shape));

/**
 * The shape of an array or a tuple.
 *
 * @template U The list of positioned element shapes, or `null` if there are no positioned elements.
 * @template R The shape of rest elements, or `null` if there are no rest elements.
 */
var ArrayShape = /** @class */ (function (_super) {
    __extends(ArrayShape, _super);
    /**
     * Creates a new {@linkcode ArrayShape} instance.
     *
     * @param shapes The list of positioned element shapes or `null` if there are no positioned elements.
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @param options The type constraint options or the type issue message.
     * @template U The list of positioned element shapes, or `null` if there are no positioned elements.
     * @template R The shape of rest elements, or `null` if there are no rest elements.
     */
    function ArrayShape(
    /**
     * The list of positioned element shapes or `null` if there are no positioned elements.
     */
    shapes, 
    /**
     * The shape of rest elements or `null` if there are no rest elements.
     */
    restShape, options) {
        var _this = _super.call(this) || this;
        _this.shapes = shapes;
        _this.restShape = restShape;
        _this._options = options;
        if (shapes !== null && restShape === null) {
            _this._typeIssueFactory = createIssueFactory(CODE_TUPLE, MESSAGE_TUPLE, options, shapes.length);
        }
        else {
            _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_ARRAY_TYPE, options, TYPE_ARRAY);
        }
        return _this;
    }
    ArrayShape.prototype.at = function (key) {
        var shapes = this.shapes;
        var index = toArrayIndex(key);
        if (index === -1) {
            return null;
        }
        if (shapes !== null && index < shapes.length) {
            return shapes[index];
        }
        return this.restShape;
    };
    /**
     * Returns an array shape that has rest elements constrained by the given shape.
     *
     * The returned object shape would have no checks.
     *
     * @param restShape The shape of rest elements or `null` if there are no rest elements.
     * @returns The new array shape.
     * @template T The shape of rest elements.
     */
    ArrayShape.prototype.rest = function (restShape) {
        return new ArrayShape(this.shapes, restShape, this._options);
    };
    /**
     * Constrains the array length.
     *
     * @param length The minimum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    ArrayShape.prototype.length = function (length, options) {
        return this.min(length, options).max(length, options);
    };
    /**
     * Constrains the minimum array length.
     *
     * @param length The minimum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    ArrayShape.prototype.min = function (length, options) {
        var issueFactory = createIssueFactory(CODE_ARRAY_MIN, MESSAGE_ARRAY_MIN, options, length);
        return setCheck(this, CODE_ARRAY_MIN, options, length, function (input, options) {
            if (input.length < length) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the maximum array length.
     *
     * @param length The maximum array length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    ArrayShape.prototype.max = function (length, options) {
        var issueFactory = createIssueFactory(CODE_ARRAY_MAX, MESSAGE_ARRAY_MAX, options, length);
        return setCheck(this, CODE_ARRAY_MAX, options, length, function (input, options) {
            if (input.length > length) {
                return issueFactory(input, options);
            }
        });
    };
    ArrayShape.prototype._requiresAsync = function () {
        return (this.shapes !== null && isAsyncShapes(this.shapes)) || (this.restShape !== null && this.restShape.async);
    };
    ArrayShape.prototype._getInputTypes = function () {
        var _a = this, shapes = _a.shapes, restShape = _a.restShape;
        var shape = shapes === null ? restShape : shapes.length === 1 ? shapes[0] : null;
        if (!this._coerced) {
            return [TYPE_ARRAY];
        }
        if (shapes === null && restShape === null) {
            // Elements aren't parsed, any value can be wrapped in an array
            return [TYPE_ANY];
        }
        if (shape === null) {
            // Iterables and array-like objects
            return [TYPE_OBJECT, TYPE_ARRAY];
        }
        return shape['_getInputTypes']().concat(TYPE_OBJECT, TYPE_ARRAY);
    };
    ArrayShape.prototype._apply = function (input, options) {
        var _a = this, shapes = _a.shapes, restShape = _a.restShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var output = input;
        var outputLength;
        var shapesLength = 0;
        var issues = null;
        // noinspection CommaExpressionJS
        if (
        // Not an array or not coercible
        (!isArray(output) && (!(options.coerced || this._coerced) || (output = this._coerce(input)) === null)) ||
            // Invalid tuple length
            ((outputLength = output.length),
                shapes !== null &&
                    (outputLength < (shapesLength = shapes.length) || (restShape === null && outputLength !== shapesLength)))) {
            return this._typeIssueFactory(input, options);
        }
        if (shapes !== null || restShape !== null) {
            for (var i = 0; i < outputLength; ++i) {
                var value = output[i];
                var valueShape = i < shapesLength ? shapes[i] : restShape;
                var result = valueShape['_apply'](value, options);
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    unshiftPath(result, i);
                    if (!options.verbose) {
                        return result;
                    }
                    issues = concatIssues(issues, result);
                    continue;
                }
                if ((_unsafe || issues === null) && !isEqual(value, result.value)) {
                    if (input === output) {
                        output = input.slice(0);
                    }
                    output[i] = result.value;
                }
            }
        }
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && input !== output) {
            return ok(output);
        }
        return issues;
    };
    ArrayShape.prototype._applyAsync = function (input, options) {
        var _this = this;
        return new Promise(function (resolve) {
            var _a = _this, shapes = _a.shapes, restShape = _a.restShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
            var output = input;
            var outputLength;
            var shapesLength = 0;
            // noinspection CommaExpressionJS
            if (
            // Not an array or not coercible
            (!isArray(output) && (!(options.coerced || _this._coerced) || (output = _this._coerce(input)) === null)) ||
                // Invalid tuple length
                ((outputLength = output.length),
                    shapes !== null &&
                        (outputLength < (shapesLength = shapes.length) || (restShape === null && outputLength !== shapesLength)))) {
                resolve(_this._typeIssueFactory(input, options));
                return;
            }
            var promises = [];
            if (shapes !== null || restShape !== null) {
                for (var i = 0; i < outputLength; ++i) {
                    var value = output[i];
                    var valueShape = i < shapesLength ? shapes[i] : restShape;
                    promises.push(valueShape['_applyAsync'](value, options));
                }
            }
            resolve(Promise.all(promises).then(function (results) {
                var resultsLength = results.length;
                var issues = null;
                for (var i = 0; i < resultsLength; ++i) {
                    var result = results[i];
                    if (result === null) {
                        continue;
                    }
                    if (isArray(result)) {
                        unshiftPath(result, i);
                        if (!options.verbose) {
                            return result;
                        }
                        issues = concatIssues(issues, result);
                        continue;
                    }
                    if ((_unsafe || issues === null) && !isEqual(input[i], result.value)) {
                        if (input === output) {
                            output = input.slice(0);
                        }
                        output[i] = result.value;
                    }
                }
                if (_applyChecks !== null && (_unsafe || issues === null)) {
                    issues = _applyChecks(output, issues, options);
                }
                if (issues === null && input !== output) {
                    return ok(output);
                }
                return issues;
            }));
        });
    };
    /**
     * Coerces value to an array or returns `null` if coercion isn't possible.
     *
     * @param value The non-array value to coerce.
     */
    ArrayShape.prototype._coerce = function (value) {
        if (isIterable(value)) {
            return Array.from(value);
        }
        return [value];
    };
    return ArrayShape;
}(CoercibleShape));

/**
 * The shape of the bigint value.
 */
var BigIntShape = /** @class */ (function (_super) {
    __extends(BigIntShape, _super);
    /**
     * Creates a new {@linkcode BigIntShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function BigIntShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_BIGINT_TYPE, options, TYPE_BIGINT);
        return _this;
    }
    BigIntShape.prototype._getInputTypes = function () {
        if (this._coerced) {
            return [TYPE_BIGINT, TYPE_STRING, TYPE_NUMBER, TYPE_BOOLEAN, TYPE_ARRAY, TYPE_UNDEFINED, TYPE_NULL];
        }
        else {
            return [TYPE_BIGINT];
        }
    };
    BigIntShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var output = input;
        var issues = null;
        var changed = false;
        if (typeof output !== 'bigint' &&
            (!(changed = options.coerced || this._coerced) || (output = this._coerce(input)) === null)) {
            return this._typeIssueFactory(input, options);
        }
        if ((_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) && changed) {
            return ok(output);
        }
        return issues;
    };
    BigIntShape.prototype._coerce = function (value) {
        if (isArray(value) && value.length === 1 && typeof (value = value[0]) === 'bigint') {
            return value;
        }
        if (value === null || value === undefined) {
            return BigInt(0);
        }
        if (typeof value === 'number' || typeof value === 'string' || typeof value === 'boolean') {
            try {
                return BigInt(value);
            }
            catch (_a) { }
        }
        return null;
    };
    return BigIntShape;
}(CoercibleShape));

/**
 * The shape of the boolean value.
 */
var BooleanShape = /** @class */ (function (_super) {
    __extends(BooleanShape, _super);
    /**
     * Creates a new {@linkcode BooleanShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function BooleanShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_BOOLEAN_TYPE, options, TYPE_BOOLEAN);
        return _this;
    }
    BooleanShape.prototype._getInputTypes = function () {
        if (this._coerced) {
            return [TYPE_BOOLEAN, TYPE_STRING, TYPE_NUMBER, TYPE_ARRAY, TYPE_NULL, TYPE_UNDEFINED];
        }
        else {
            return [TYPE_BOOLEAN];
        }
    };
    BooleanShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var output = input;
        var issues = null;
        var changed = false;
        if (typeof output !== 'boolean' &&
            (!(changed = options.coerced || this._coerced) || (output = this._coerce(input)) === null)) {
            return this._typeIssueFactory(input, options);
        }
        if ((_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) && changed) {
            return ok(output);
        }
        return issues;
    };
    /**
     * Coerces value to a boolean or returns `null` if coercion isn't possible.
     *
     * @param value The non-boolean value to coerce.
     */
    BooleanShape.prototype._coerce = function (value) {
        if (isArray(value) && value.length === 1 && typeof (value = value[0]) === 'boolean') {
            return value;
        }
        if (value === null || value === undefined || value === false || value === 0 || value === 'false') {
            return false;
        }
        if (value === true || value === 1 || value === 'true') {
            return true;
        }
        return null;
    };
    return BooleanShape;
}(CoercibleShape));

/**
 * The shape that constrains an input to be the exact value.
 *
 * @template T The value.
 */
var ConstShape = /** @class */ (function (_super) {
    __extends(ConstShape, _super);
    /**
     * Creates a new {@linkcode ConstShape} instance.
     *
     * @param value The exact value.
     * @param options The type constraint options or an issue message.
     * @template T Allowed values.
     */
    function ConstShape(value, options) {
        var _this = _super.call(this) || this;
        _this.value = value;
        _this._typePredicate = value !== value ? Number.isNaN : function (input) { return value === input; };
        _this._typeIssueFactory = createIssueFactory(CODE_CONST, MESSAGE_CONST, options, value);
        return _this;
    }
    ConstShape.prototype._getInputTypes = function () {
        return [getValueType(this.value)];
    };
    ConstShape.prototype._getInputValues = function () {
        return [this.value];
    };
    ConstShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        if (!this._typePredicate(input)) {
            return this._typeIssueFactory(input, options);
        }
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    return ConstShape;
}(Shape));

/**
 * The shape of the `Date` object.
 */
var DateShape = /** @class */ (function (_super) {
    __extends(DateShape, _super);
    /**
     * Creates a new {@linkcode DateShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function DateShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_DATE_TYPE, options, TYPE_DATE);
        return _this;
    }
    DateShape.prototype._getInputTypes = function () {
        if (this._coerced) {
            return [TYPE_DATE, TYPE_STRING, TYPE_NUMBER, TYPE_ARRAY];
        }
        else {
            return [TYPE_DATE];
        }
    };
    DateShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var output = input;
        var issues = null;
        var changed = false;
        if (!isDate(input) && (!(changed = options.coerced || this._coerced) || (output = this._coerce(input)) === null)) {
            return this._typeIssueFactory(input, options);
        }
        if ((_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) && changed) {
            return ok(output);
        }
        return issues;
    };
    /**
     * Coerces value to a `Date` or returns `null` if coercion isn't possible.
     *
     * @param value The non-`Date` value to coerce.
     */
    DateShape.prototype._coerce = function (value) {
        if (isArray(value) && value.length === 1 && isDate((value = value[0]))) {
            return value;
        }
        if ((typeof value === 'string' || typeof value === 'number') && isDate((value = new Date(value)))) {
            return value;
        }
        return null;
    };
    return DateShape;
}(CoercibleShape));

var UNRECOGNIZED = Symbol();
/**
 * The shape that constrains an input to one of values.
 *
 * @template T Allowed values.
 */
var EnumShape = /** @class */ (function (_super) {
    __extends(EnumShape, _super);
    /**
     * Creates a new {@linkcode EnumShape} instance.
     *
     * @param source The list of allowed values, a const key-value mapping, or an enum object.
     * @param options The type constraint options or an issue message.
     * @template T Allowed values.
     */
    function EnumShape(
    /**
     * The list of allowed values, a const key-value mapping, or an enum object.
     */
    source, options) {
        var _this = _super.call(this) || this;
        _this.source = source;
        var valueMapping;
        var values;
        if (isArray(source)) {
            valueMapping = null;
            values = unique(source);
        }
        else {
            valueMapping = source;
            values = unique(getEnumValues(source));
        }
        _this.values = values;
        _this._valueMapping = valueMapping;
        _this._typeIssueFactory = createIssueFactory(CODE_ENUM, MESSAGE_ENUM, options, values);
        return _this;
    }
    EnumShape.prototype._getInputTypes = function () {
        var valueTypes = this.values.map(getValueType);
        if (this._coerced) {
            return valueTypes.concat(TYPE_STRING, TYPE_ARRAY);
        }
        else {
            return valueTypes;
        }
    };
    EnumShape.prototype._getInputValues = function () {
        return this.values.slice(0);
    };
    EnumShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var output = input;
        var issues = null;
        var changed = false;
        if (!this.values.includes(output) &&
            (!(changed = options.coerced || this._coerced) || (output = this._coerce(input)) === UNRECOGNIZED)) {
            return this._typeIssueFactory(input, options);
        }
        if ((_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) && changed) {
            return ok(output);
        }
        return issues;
    };
    /**
     * Coerces value to an enum value or returns {@linkcode UNRECOGNIZED} if coercion isn't possible.
     *
     * @param value The non-enum value to coerce.
     */
    EnumShape.prototype._coerce = function (value) {
        var _valueMapping = this._valueMapping;
        if (isArray(value) && value.length === 1 && this.values.includes((value = value[0]))) {
            return value;
        }
        if (_valueMapping !== null && typeof value === 'string' && value in _valueMapping) {
            return _valueMapping[value];
        }
        return UNRECOGNIZED;
    };
    return EnumShape;
}(CoercibleShape));
/**
 * Returns values of the enum. Source must contain key-value and value-key mapping to be considered a native enum.
 */
function getEnumValues(source) {
    var values = [];
    for (var key in source) {
        var a = source[key];
        var b = source[a];
        var aType = typeof a;
        var bType = typeof b;
        if (((aType !== 'string' || bType !== 'number') && (aType !== 'number' || bType !== 'string')) || b != key) {
            return Object.values(source);
        }
        if (typeof a === 'number') {
            values.push(a);
        }
    }
    return values;
}

/**
 * The shape of the class instance.
 *
 * @template C The class constructor.
 */
var InstanceShape = /** @class */ (function (_super) {
    __extends(InstanceShape, _super);
    /**
     * Creates a new {@linkcode InstanceShape} instance.
     *
     * @param ctor The class constructor.
     * @param options The type constraint options or the type issue message.
     * @template C The class constructor.
     */
    function InstanceShape(ctor, options) {
        var _this = _super.call(this) || this;
        _this.ctor = ctor;
        _this._typeIssueFactory = createIssueFactory(CODE_INSTANCE, MESSAGE_INSTANCE, options, ctor);
        return _this;
    }
    InstanceShape.prototype._getInputTypes = function () {
        if (this.ctor === Array || Array.prototype.isPrototypeOf(this.ctor.prototype)) {
            return [TYPE_ARRAY];
        }
        else {
            return [TYPE_OBJECT];
        }
    };
    InstanceShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        if (!(input instanceof this.ctor)) {
            return this._typeIssueFactory(input, options);
        }
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    return InstanceShape;
}(Shape));

var IntersectionShape = /** @class */ (function (_super) {
    __extends(IntersectionShape, _super);
    function IntersectionShape(shapes, options) {
        var _this = _super.call(this) || this;
        _this.shapes = shapes;
        _this._typeIssueFactory = createIssueFactory(CODE_INTERSECTION, MESSAGE_INTERSECTION, options, undefined);
        return _this;
    }
    IntersectionShape.prototype._requiresAsync = function () {
        return isAsyncShapes(this.shapes);
    };
    IntersectionShape.prototype._getInputTypes = function () {
        return intersectValueTypes(this.shapes.map(function (shape) { return shape['_getInputTypes'](); }));
    };
    IntersectionShape.prototype._apply = function (input, options) {
        var shapes = this.shapes;
        var shapesLength = shapes.length;
        var outputs = null;
        for (var i = 0; i < shapesLength; ++i) {
            var result = shapes[i]['_apply'](input, options);
            if (result === null) {
                continue;
            }
            if (isArray(result)) {
                return result;
            }
            var output = result.value;
            if (outputs === null) {
                outputs = [output];
                continue;
            }
            if (outputs.includes(output)) {
                continue;
            }
            outputs.push(output);
        }
        if (outputs === null) {
            return null;
        }
        if (outputs.length !== shapesLength) {
            outputs.push(input);
        }
        return intersectOutputs(input, outputs, this._typeIssueFactory, options);
    };
    IntersectionShape.prototype._applyAsync = function (input, options) {
        var _this = this;
        var shapes = this.shapes;
        var shapesLength = shapes.length;
        var promises = [];
        for (var i = 0; i < shapesLength; ++i) {
            promises.push(shapes[i]['_applyAsync'](input, options));
        }
        return Promise.all(promises).then(function (results) {
            var outputs = null;
            for (var i = 0; i < shapesLength; ++i) {
                var result = results[i];
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    return result;
                }
                var output = result.value;
                if (outputs === null) {
                    outputs = [output];
                    continue;
                }
                if (outputs.includes(output)) {
                    continue;
                }
                outputs.push(output);
            }
            if (outputs === null) {
                return null;
            }
            if (outputs.length !== shapesLength) {
                outputs.push(input);
            }
            return intersectOutputs(input, outputs, _this._typeIssueFactory, options);
        });
    };
    return IntersectionShape;
}(Shape));
var NEVER = Symbol();
function intersectOutputs(input, outputs, issueFactory, options) {
    var outputsLength = outputs.length;
    var value = outputs[0];
    for (var i = 1; i < outputsLength; ++i) {
        value = intersectValues(value, outputs[i]);
        if (value === NEVER) {
            return issueFactory(input, options);
        }
    }
    return ok(value);
}
function intersectValues(a, b) {
    if (isEqual(a, b)) {
        return a;
    }
    var aType = getValueType(a);
    var bType = getValueType(b);
    if (aType !== bType) {
        return NEVER;
    }
    if (aType === TYPE_DATE) {
        if (a.getTime() === b.getTime()) {
            return a;
        }
        return NEVER;
    }
    if (aType === TYPE_OBJECT) {
        var output = Object.assign({}, a, b);
        for (var key in a) {
            if (key in b) {
                var outputValue = intersectValues(a[key], b[key]);
                if (outputValue === NEVER) {
                    return NEVER;
                }
                output[key] = outputValue;
            }
        }
        return output;
    }
    if (aType === TYPE_ARRAY) {
        var aLength = a.length;
        if (aLength !== b.length) {
            return NEVER;
        }
        var output = a;
        for (var i = 0; i < aLength; ++i) {
            var aValue = a[i];
            var bValue = b[i];
            if (isEqual(aValue, bValue)) {
                continue;
            }
            if (output === a) {
                output = a.slice(0);
            }
            var outputValue = intersectValues(aValue, bValue);
            if (outputValue === NEVER) {
                return NEVER;
            }
            output[i] = outputValue;
        }
        return output;
    }
    return NEVER;
}
function intersectValueTypes(arr) {
    var arrLength = arr.length;
    if (arrLength === 0) {
        return [TYPE_NEVER];
    }
    var types = arr[0];
    for (var i = 1; i < arr.length; ++i) {
        for (var j = 0; j < types.length; ++j) {
            if (arr[i].includes(types[j])) {
                continue;
            }
            types.splice(j--, 1);
        }
    }
    if (types.length === 0) {
        return [TYPE_NEVER];
    }
    return types;
}

/**
 * The shape of a value deserialized from a JSON string.
 */
var JSONShape = /** @class */ (function (_super) {
    __extends(JSONShape, _super);
    /**
     * Creates a new {@linkcode JSONShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function JSONShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_STRING_TYPE, options, TYPE_STRING);
        _this._jsonIssueFactory = createIssueFactory(CODE_JSON, function (param) { return param; }, options);
        return _this;
    }
    /**
     * Prescribes how each value originally produced by parsing is transformed before being returned.
     *
     * @param reviver The reviver callback.
     * @returns The clone of the shape.
     */
    JSONShape.prototype.revive = function (
    /**
     * Prescribes how each value originally produced by parsing is transformed before being returned.
     *
     * @param key The key associated with the value.
     * @param value The value produced by parsing.
     * @returns The transformed value.
     */
    reviver) {
        var shape = cloneObject(this);
        shape._reviver = reviver;
        return shape;
    };
    JSONShape.prototype._getInputTypes = function () {
        return [TYPE_STRING];
    };
    JSONShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var output;
        var issues;
        if (typeof input !== 'string') {
            return this._typeIssueFactory(input, options);
        }
        try {
            output = JSON.parse(input, this._reviver);
        }
        catch (error) {
            return this._jsonIssueFactory(input, options, error.message);
        }
        if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
            return ok(output);
        }
        return issues;
    };
    return JSONShape;
}(Shape));

/**
 * Lazily resolves a shape using the provider.
 *
 * @template S The base shape.
 */
var LazyShape = /** @class */ (function (_super) {
    __extends(LazyShape, _super);
    /**
     * Creates a new {@linkcode LazyShape} instance.
     *
     * @param shapeProvider The provider that returns the base shape.
     * @template S The base shape.
     */
    function LazyShape(shapeProvider) {
        var _this = _super.call(this) || this;
        _this._shapeProvider = shapeProvider;
        return _this;
    }
    Object.defineProperty(LazyShape.prototype, "shape", {
        /**
         * The base shape.
         */
        get: function () {
            var shape = this._shapeProvider();
            if (!(shape instanceof Shape)) {
                throw new Error(ERROR_SHAPE_EXPECTED);
            }
            Object.defineProperty(this, 'shape', { value: shape });
            return shape;
        },
        enumerable: false,
        configurable: true
    });
    LazyShape.prototype._requiresAsync = function () {
        var _requiresAsync = this._requiresAsync;
        this._requiresAsync = returnFalse;
        try {
            return this.shape.async;
        }
        finally {
            this._requiresAsync = _requiresAsync;
        }
    };
    LazyShape.prototype._getInputTypes = function () {
        var _getInputTypes = this._getInputTypes;
        this._getInputTypes = returnArray;
        try {
            return this.shape['_getInputTypes']();
        }
        finally {
            this._getInputTypes = _getInputTypes;
        }
    };
    LazyShape.prototype._getInputValues = function () {
        var _getInputValues = this._getInputValues;
        this._getInputValues = returnArray;
        try {
            return this.shape['_getInputValues']();
        }
        finally {
            this._getInputValues = _getInputValues;
        }
    };
    LazyShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var output = input;
        var issues;
        var result = this.shape['_apply'](input, options);
        if (result !== null) {
            if (isArray(result)) {
                return result;
            }
            output = result.value;
        }
        if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
            return result;
        }
        return issues;
    };
    LazyShape.prototype._applyAsync = function (input, options) {
        var _applyChecks = this._applyChecks;
        return this.shape['_applyAsync'](input, options).then(function (result) {
            var output = input;
            var issues;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                output = result.value;
            }
            if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
                return result;
            }
            return issues;
        });
    };
    return LazyShape;
}(Shape));

/**
 * The shape of a `Map` instance.
 *
 * @template K The key shape.
 * @template V The value shape.
 */
var MapShape = /** @class */ (function (_super) {
    __extends(MapShape, _super);
    /**
     * Creates a new {@linkcode MapShape} instance.
     *
     * @param keyShape The key shape.
     * @param valueShape The value shape.
     * @param options The type constraint options or an issue message.
     * @template K The key shape.
     * @template V The value shape.
     */
    function MapShape(
    /**
     * The key shape.
     */
    keyShape, 
    /**
     * The value shape.
     */
    valueShape, options) {
        var _this = _super.call(this) || this;
        _this.keyShape = keyShape;
        _this.valueShape = valueShape;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_MAP_TYPE, options, TYPE_MAP);
        return _this;
    }
    MapShape.prototype.at = function (key) {
        return this.valueShape;
    };
    MapShape.prototype._requiresAsync = function () {
        return this.keyShape.async || this.valueShape.async;
    };
    MapShape.prototype._getInputTypes = function () {
        if (this._coerced) {
            return [TYPE_OBJECT, TYPE_ARRAY];
        }
        else {
            return [TYPE_OBJECT];
        }
    };
    MapShape.prototype._apply = function (input, options) {
        var changed = false;
        var entries;
        if (
        // Not a Map
        !(input instanceof Map && (entries = Array.from(input.entries()))) &&
            // No coercion or not coercible
            (!(options.coerced || this._coerced) || !(changed = (entries = this._coerceEntries(input)) !== null))) {
            return this._typeIssueFactory(input, options);
        }
        var _a = this, keyShape = _a.keyShape, valueShape = _a.valueShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var entriesLength = entries.length;
        var issues = null;
        for (var i = 0; i < entriesLength; ++i) {
            var _b = entries[i], key = _b[0], value = _b[1];
            var outputKey = key;
            var outputValue = value;
            var keyResult = keyShape['_apply'](key, options);
            if (keyResult !== null) {
                if (isArray(keyResult)) {
                    unshiftPath(keyResult, key);
                    if (!options.verbose) {
                        return keyResult;
                    }
                    issues = concatIssues(issues, keyResult);
                }
                else {
                    outputKey = keyResult.value;
                }
            }
            var valueResult = valueShape['_apply'](value, options);
            if (valueResult !== null) {
                if (isArray(valueResult)) {
                    unshiftPath(valueResult, key);
                    if (!options.verbose) {
                        return valueResult;
                    }
                    issues = concatIssues(issues, valueResult);
                }
                else {
                    outputValue = valueResult.value;
                }
            }
            if ((_unsafe || issues === null) && (!isEqual(key, outputKey) || !isEqual(value, outputValue))) {
                changed = true;
                entries[i] = [outputKey, outputValue];
            }
        }
        var output = changed ? new Map(entries) : input;
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (changed && issues === null) {
            return ok(output);
        }
        return issues;
    };
    MapShape.prototype._applyAsync = function (input, options) {
        var _this = this;
        return new Promise(function (resolve) {
            var changed = false;
            var entries;
            if (
            // Not a Map
            !(input instanceof Map && (entries = Array.from(input.entries()))) &&
                // No coercion or not coercible
                (!(options.coerced || _this._coerced) || !(changed = (entries = _this._coerceEntries(input)) !== null))) {
                resolve(_this._typeIssueFactory(input, options));
                return;
            }
            var _a = _this, keyShape = _a.keyShape, valueShape = _a.valueShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
            var entriesLength = entries.length;
            var promises = [];
            for (var i = 0; i < entriesLength; ++i) {
                var _b = entries[i], key = _b[0], value = _b[1];
                promises.push(keyShape['_applyAsync'](key, options), valueShape['_applyAsync'](value, options));
            }
            resolve(Promise.all(promises).then(function (results) {
                var issues = null;
                for (var i = 0; i < entriesLength; ++i) {
                    var _a = entries[i], key = _a[0], value = _a[1];
                    var outputKey = key;
                    var outputValue = value;
                    var keyResult = results[i * 2];
                    var valueResult = results[i * 2 + 1];
                    if (keyResult !== null) {
                        if (isArray(keyResult)) {
                            unshiftPath(keyResult, key);
                            if (!options.verbose) {
                                return keyResult;
                            }
                            issues = concatIssues(issues, keyResult);
                        }
                        else {
                            outputKey = keyResult.value;
                        }
                    }
                    if (valueResult !== null) {
                        if (isArray(valueResult)) {
                            unshiftPath(valueResult, key);
                            if (!options.verbose) {
                                return valueResult;
                            }
                            issues = concatIssues(issues, valueResult);
                        }
                        else {
                            outputValue = valueResult.value;
                        }
                    }
                    if ((_unsafe || issues === null) && (!isEqual(key, outputKey) || !isEqual(value, outputValue))) {
                        changed = true;
                        entries[i] = [outputKey, outputValue];
                    }
                }
                var output = changed ? new Map(entries) : input;
                if (_applyChecks !== null && (_unsafe || issues === null)) {
                    issues = _applyChecks(output, issues, options);
                }
                if (changed && issues === null) {
                    return ok(output);
                }
                return issues;
            }));
        });
    };
    /**
     * Coerces value to a list of `Map` entries, or returns `null` if coercion isn't possible.
     *
     * @param value The non-`Map` value to coerce.
     */
    MapShape.prototype._coerceEntries = function (value) {
        if (isArray(value)) {
            return value.every(isEntry) ? value : null;
        }
        if (isIterable(value)) {
            value = Array.from(value);
            return value.every(isEntry) ? value : null;
        }
        if (isObjectLike(value)) {
            return Object.entries(value);
        }
        return null;
    };
    return MapShape;
}(CoercibleShape));
function isEntry(value) {
    return isArray(value) && value.length === 2;
}

/**
 * The shape that doesn't match any input.
 */
var NeverShape = /** @class */ (function (_super) {
    __extends(NeverShape, _super);
    /**
     * Creates a new {@linkcode NeverShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function NeverShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_NEVER_TYPE, options, TYPE_NEVER);
        return _this;
    }
    NeverShape.prototype._getInputTypes = function () {
        return [TYPE_NEVER];
    };
    NeverShape.prototype._apply = function (input, options) {
        return this._typeIssueFactory(input, options);
    };
    return NeverShape;
}(Shape));

/**
 * The shape that constrains the input as a number.
 */
var NumberShape = /** @class */ (function (_super) {
    __extends(NumberShape, _super);
    /**
     * Creates a new {@linkcode NumberShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function NumberShape(options) {
        var _this = _super.call(this) || this;
        _this._typePredicate = isNumber;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_NUMBER_TYPE, options, TYPE_NUMBER);
        return _this;
    }
    /**
     * Constrains the number to be greater than zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.positive = function (options) {
        return this.gt(0, options);
    };
    /**
     * Constrains the number to be less than zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.negative = function (options) {
        return this.lt(0, options);
    };
    /**
     * Constrains the number to be less or equal to zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.nonPositive = function (options) {
        return this.lte(0, options);
    };
    /**
     * Constrains the number to be greater or equal to zero.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.nonNegative = function (options) {
        return this.gte(0, options);
    };
    /**
     * Constrains the number to be greater than the value.
     *
     * @param value The exclusive minimum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.gt = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_GT, MESSAGE_NUMBER_GT, options, value);
        return setCheck(this.deleteCheck(CODE_NUMBER_GTE), CODE_NUMBER_GT, options, value, function (input, options) {
            if (input <= value) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be less than the value.
     *
     * @param value The exclusive maximum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.lt = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_LT, MESSAGE_NUMBER_LT, options, value);
        return setCheck(this.deleteCheck(CODE_NUMBER_LTE), CODE_NUMBER_LT, options, value, function (input, options) {
            if (input >= value) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be greater than or equal to the value.
     *
     * @param value The inclusive minimum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.gte = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_GTE, MESSAGE_NUMBER_GTE, options, value);
        return setCheck(this.deleteCheck(CODE_NUMBER_GT), CODE_NUMBER_GTE, options, value, function (input, options) {
            if (input < value) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be less than or equal to the value.
     *
     * @param value The inclusive maximum value.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.lte = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_LTE, MESSAGE_NUMBER_LTE, options, value);
        return setCheck(this.deleteCheck(CODE_NUMBER_LT), CODE_NUMBER_LTE, options, value, function (input, options) {
            if (input > value) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be a multiple of the divisor.
     *
     * @param value The positive number by which the input should be divisible without a remainder.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.multipleOf = function (value, options) {
        var issueFactory = createIssueFactory(CODE_NUMBER_MULTIPLE_OF, MESSAGE_NUMBER_MULTIPLE_OF, options, value);
        return setCheck(this, CODE_NUMBER_MULTIPLE_OF, options, value, function (input, options) {
            if (!isMultipleOf(input, value)) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the number to be a finite number.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.finite = function (options) {
        var shape = cloneObject(this);
        shape._typeIssueFactory = createIssueFactory(CODE_NUMBER_FINITE, MESSAGE_NUMBER_FINITE, options, undefined);
        shape._typePredicate = Number.isFinite;
        return shape;
    };
    /**
     * Constrains the number to be an integer.
     *
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    NumberShape.prototype.integer = function (options) {
        var shape = cloneObject(this);
        shape._typeIssueFactory = createIssueFactory(CODE_NUMBER_INTEGER, MESSAGE_NUMBER_INTEGER, options, undefined);
        shape._typePredicate = Number.isInteger;
        return shape;
    };
    /**
     * Allows `NaN` as an input and output value, or replaces an input `NaN` value with a default output value.
     *
     * @param [defaultValue = NaN] The value that is used instead of `NaN` in the output.
     */
    NumberShape.prototype.nan = function (defaultValue) {
        if (defaultValue === void 0) { defaultValue = NaN; }
        return this.replace(NaN, defaultValue);
    };
    NumberShape.prototype._getInputTypes = function () {
        if (this._coerced) {
            return [TYPE_NUMBER, TYPE_STRING, TYPE_BOOLEAN, TYPE_ARRAY, TYPE_DATE, TYPE_UNDEFINED, TYPE_NULL];
        }
        else {
            return [TYPE_NUMBER];
        }
    };
    NumberShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var output = input;
        var issues = null;
        var changed = false;
        if (!this._typePredicate(output) &&
            (!(changed = options.coerced || this._coerced) || (output = this._coerce(input)) === null)) {
            return this._typeIssueFactory(input, options);
        }
        if ((_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) && changed) {
            return ok(output);
        }
        return issues;
    };
    /**
     * Coerces value to a number (not `NaN`) or returns `null` if coercion isn't possible.
     *
     * @param value The non-number value to coerce.
     */
    NumberShape.prototype._coerce = function (value) {
        if (isArray(value) && value.length === 1 && typeof (value = value[0]) === 'number' && value === value) {
            return value;
        }
        if (value === null || value === undefined) {
            return 0;
        }
        if ((typeof value === 'string' || typeof value === 'boolean' || value instanceof Date) &&
            this._typePredicate((value = +value))) {
            return value;
        }
        return null;
    };
    return NumberShape;
}(CoercibleShape));
NumberShape.prototype.min = NumberShape.prototype.gte;
NumberShape.prototype.max = NumberShape.prototype.lte;
/**
 * Checks that `a` is divisible without a remainder by `b`.
 */
function isMultipleOf(a, b) {
    if (b <= 0 || a !== a || b !== b || a === Infinity || a === -Infinity || b === Infinity) {
        return false;
    }
    if (a === b || a % b === 0) {
        return true;
    }
    if (Math.abs(a) < b || a === Math.floor(a) || b === Math.floor(b)) {
        // Integers had their chance
        return false;
    }
    var aStr = a.toString();
    var bStr = b.toString();
    var aExpIndex = aStr.indexOf('e');
    var bExpIndex = bStr.indexOf('e');
    var aDotIndex = aStr.indexOf('.');
    var bDotIndex = bStr.indexOf('.');
    // The exponent extracted from e+XXX plus the number of decimal digits
    var aDecLength = (~aExpIndex && Math.abs(+aStr.substring(aExpIndex + 1))) +
        (~aDotIndex && ~(~aExpIndex || ~aStr.length) + ~aDotIndex);
    var bDecLength = (~bExpIndex && Math.abs(+bStr.substring(bExpIndex + 1))) +
        (~bDotIndex && ~(~bExpIndex || ~bStr.length) + ~bDotIndex);
    var decLength = Math.max(aDecLength, bDecLength);
    // Undefined behaviour if long overflow occurs
    var aLong = +a.toFixed(decLength).replace('.', '');
    var bLong = +b.toFixed(decLength).replace('.', '');
    return aLong % bLong === 0;
}

/**
 * The shape of an object.
 *
 * @template P The mapping from an object key to a corresponding value shape.
 * @template R The shape that constrains values of
 * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
 */
var ObjectShape = /** @class */ (function (_super) {
    __extends(ObjectShape, _super);
    /**
     * Creates a new {@linkcode ObjectShape} instance.
     *
     * @param shapes The mapping from an object key to a corresponding value shape.
     * @param restShape The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures). If `null`
     * then values thea fall under the index signature are unconstrained.
     * @param options The type constraint options or an issue message.
     * @param keysMode The mode of unknown keys handling.
     * @template P The mapping from an object key to a corresponding value shape.
     * @template R The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    function ObjectShape(
    /**
     * The mapping from an object key to a corresponding value shape.
     */
    shapes, 
    /**
     * The shape that constrains values of
     * [a string index signature](https://www.typescriptlang.org/docs/handbook/2/objects.html#index-signatures).
     */
    restShape, options, keysMode) {
        if (keysMode === void 0) { keysMode = 'preserved'; }
        var _this = _super.call(this) || this;
        _this.shapes = shapes;
        _this.restShape = restShape;
        _this._typePredicate = isObjectLike;
        _this.keys = Object.keys(shapes);
        _this.keysMode = keysMode;
        _this._options = options;
        _this._valueShapes = Object.values(shapes);
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_OBJECT_TYPE, options, TYPE_OBJECT);
        return _this;
    }
    ObjectShape.prototype.at = function (key) {
        return this.shapes.hasOwnProperty(key) ? this.shapes[key] : this.restShape;
    };
    ObjectShape.prototype.extend = function (shape) {
        var shapes = Object.assign({}, this.shapes, shape instanceof ObjectShape ? shape.shapes : shape);
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    /**
     * Returns an object shape that only has properties with listed keys.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to pick.
     * @returns The new object shape.
     * @template K The tuple of keys to pick.
     */
    ObjectShape.prototype.pick = function (keys) {
        var shapes = {};
        for (var key in this.shapes) {
            if (keys.includes(key)) {
                shapes[key] = this.shapes[key];
            }
        }
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    /**
     * Returns an object shape that doesn't have the listed keys.
     *
     * The returned object shape would have no checks.
     *
     * @param keys The list of property keys to omit.
     * @returns The new object shape.
     * @template K The tuple of keys to omit.
     */
    ObjectShape.prototype.omit = function (keys) {
        var shapes = {};
        for (var key in this.shapes) {
            if (!keys.includes(key)) {
                shapes[key] = this.shapes[key];
            }
        }
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    ObjectShape.prototype.partial = function (keys) {
        var shapes = {};
        for (var key in this.shapes) {
            var shape = this.shapes[key];
            if (keys === undefined || keys.includes(key)) {
                shape = shape.optional();
            }
            shapes[key] = shape;
        }
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    ObjectShape.prototype.required = function (keys) {
        var shapes = {};
        for (var key in this.shapes) {
            var shape = this.shapes[key];
            if (keys === undefined || keys.includes(key)) {
                shape = shape.nonOptional();
            }
            shapes[key] = shape;
        }
        return new ObjectShape(shapes, this.restShape, this._options, this.keysMode);
    };
    /**
     * Returns an object shape that allows only known keys and has no index signature.
     *
     * The returned object shape would have no checks.
     *
     * @param options The constraint options or an issue message.
     * @returns The new object shape.
     */
    ObjectShape.prototype.exact = function (options) {
        var shape = new ObjectShape(this.shapes, null, this._options, 'exact');
        shape._exactIssueFactory = createIssueFactory(CODE_UNKNOWN_KEYS, MESSAGE_UNKNOWN_KEYS, options);
        return shape;
    };
    /**
     * Returns an object shape that doesn't have an index signature and all unknown keys are stripped.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    ObjectShape.prototype.strip = function () {
        return new ObjectShape(this.shapes, null, this._options, 'stripped');
    };
    /**
     * Returns an object shape that has an index signature that doesn't constrain values.
     *
     * The returned object shape would have no checks.
     *
     * @returns The new object shape.
     */
    ObjectShape.prototype.preserve = function () {
        return new ObjectShape(this.shapes, null, this._options);
    };
    /**
     * Returns an object shape that has an index signature that is constrained by the given shape.
     *
     * The returned object shape would have no checks.
     *
     * @param restShape The shape that validates values at unknown keys.
     * @returns The new object shape.
     * @template T The index signature shape.
     */
    ObjectShape.prototype.rest = function (restShape) {
        return new ObjectShape(this.shapes, restShape, this._options);
    };
    /**
     * Returns the enum shape of keys of this object.
     */
    ObjectShape.prototype.keyof = function () {
        return new EnumShape(this.keys);
    };
    /**
     * Constrains an object to be an `Object` instance or to have a `null` prototype.
     */
    ObjectShape.prototype.plain = function () {
        var shape = cloneObject(this);
        shape._typePredicate = isPlainObject;
        return shape;
    };
    ObjectShape.prototype._requiresAsync = function () {
        return (this.restShape !== null && this.restShape.async) || isAsyncShapes(Object.values(this.shapes));
    };
    ObjectShape.prototype._getInputTypes = function () {
        return [TYPE_OBJECT];
    };
    ObjectShape.prototype._apply = function (input, options) {
        if (!this._typePredicate(input)) {
            return this._typeIssueFactory(input, options);
        }
        if (this.keysMode === 'preserved' && this.restShape === null) {
            return this._applyRestUnchecked(input, options);
        }
        else {
            return this._applyRestChecked(input, options);
        }
    };
    ObjectShape.prototype._applyAsync = function (input, options) {
        var _this = this;
        return new Promise(function (resolve) {
            if (!_this._typePredicate(input)) {
                resolve(_this._typeIssueFactory(input, options));
                return;
            }
            var _a = _this, keys = _a.keys, keysMode = _a.keysMode, restShape = _a.restShape, _valueShapes = _a._valueShapes, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
            var keysLength = keys.length;
            var promises = [];
            var issues = null;
            var output = input;
            var seenCount = 0;
            var seenBits = 0;
            var unknownKeys = null;
            for (var key in input) {
                var value = input[key];
                var index = keys.indexOf(key);
                var valueShape = restShape;
                if (index !== -1) {
                    seenCount++;
                    seenBits = enableBitAt(seenBits, index);
                    valueShape = _valueShapes[index];
                }
                if (valueShape !== null) {
                    promises.push(key, valueShape['_applyAsync'](value, options));
                    continue;
                }
                if (keysMode === 'exact') {
                    if (unknownKeys !== null) {
                        unknownKeys.push(key);
                        continue;
                    }
                    unknownKeys = [key];
                    if (!options.verbose) {
                        break;
                    }
                    continue;
                }
                if (input === output && keysMode === 'stripped') {
                    output = cloneObjectKnownKeys(input, keys);
                }
            }
            if (unknownKeys !== null) {
                var issue = _this._exactIssueFactory(input, options, unknownKeys);
                if (!options.verbose) {
                    resolve(issue);
                    return;
                }
                issues = concatIssues(issues, issue);
            }
            if (seenCount !== keysLength) {
                for (var i = 0; i < keysLength; ++i) {
                    if (isBitEnabledAt(seenBits, i)) {
                        continue;
                    }
                    var key = keys[i];
                    var value = input[key];
                    promises.push(key, _valueShapes[i]['_applyAsync'](value, options));
                }
            }
            resolve(Promise.all(promises).then(function (entries) {
                var entriesLength = entries.length;
                for (var i = 0; i < entriesLength; i += 2) {
                    var key = entries[i];
                    var result = entries[i + 1];
                    if (result === null) {
                        continue;
                    }
                    if (isArray(result)) {
                        unshiftPath(result, key);
                        if (!options.verbose) {
                            return result;
                        }
                        issues = concatIssues(issues, result);
                        continue;
                    }
                    if ((_unsafe || issues === null) && !isEqual(input[key], result.value)) {
                        if (input === output) {
                            output = cloneObjectEnumerableKeys(input);
                        }
                        setKeyValue(output, key, result.value);
                    }
                }
                if (_applyChecks !== null && (_unsafe || issues === null)) {
                    issues = _applyChecks(output, issues, options);
                }
                if (issues === null && input !== output) {
                    return ok(output);
                }
                return issues;
            }));
        });
    };
    /**
     * Unknown keys are preserved as is and aren't checked.
     */
    ObjectShape.prototype._applyRestUnchecked = function (input, options) {
        var _a = this, keys = _a.keys, _valueShapes = _a._valueShapes, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var keysLength = keys.length;
        var issues = null;
        var output = input;
        for (var i = 0; i < keysLength; ++i) {
            var key = keys[i];
            var value = input[key];
            var result = _valueShapes[i]['_apply'](value, options);
            if (result === null) {
                continue;
            }
            if (isArray(result)) {
                unshiftPath(result, key);
                if (!options.verbose) {
                    return result;
                }
                issues = concatIssues(issues, result);
                continue;
            }
            if ((_unsafe || issues === null) && !isEqual(value, result.value)) {
                if (input === output) {
                    output = cloneObjectEnumerableKeys(input);
                }
                setKeyValue(output, key, result.value);
            }
        }
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && input !== output) {
            return ok(output);
        }
        return issues;
    };
    /**
     * Unknown keys are either parsed with a {@linkcode restShape}, stripped, or cause an issue.
     */
    ObjectShape.prototype._applyRestChecked = function (input, options) {
        var _a = this, keys = _a.keys, keysMode = _a.keysMode, restShape = _a.restShape, _valueShapes = _a._valueShapes, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var keysLength = keys.length;
        var issues = null;
        var output = input;
        var seenCount = 0;
        var seenBits = 0;
        var unknownKeys = null;
        for (var key in input) {
            var value = input[key];
            var index = keys.indexOf(key);
            var valueShape = restShape;
            // The key is known
            if (index !== -1) {
                seenCount++;
                seenBits = enableBitAt(seenBits, index);
                valueShape = _valueShapes[index];
            }
            // The key is known or indexed
            if (valueShape !== null) {
                var result = valueShape['_apply'](value, options);
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    unshiftPath(result, key);
                    if (!options.verbose) {
                        return result;
                    }
                    issues = concatIssues(issues, result);
                    continue;
                }
                if ((_unsafe || issues === null) && !isEqual(value, result.value)) {
                    if (input === output) {
                        output = restShape === null ? cloneObjectKnownKeys(input, keys) : cloneObjectEnumerableKeys(input);
                    }
                    setKeyValue(output, key, result.value);
                }
                continue;
            }
            // Unknown keys raise an issue
            if (keysMode === 'exact') {
                if (unknownKeys !== null) {
                    unknownKeys.push(key);
                    continue;
                }
                unknownKeys = [key];
                if (!options.verbose) {
                    break;
                }
                continue;
            }
            // Unknown keys are stripped
            if (input === output && (_unsafe || issues === null)) {
                output = cloneObjectKnownKeys(input, keys);
            }
        }
        // Raise unknown keys issue
        if (unknownKeys !== null) {
            var issue = this._exactIssueFactory(input, options, unknownKeys);
            if (!options.verbose) {
                return issue;
            }
            issues = concatIssues(issues, issue);
        }
        // Parse absent known keys
        if (seenCount !== keysLength) {
            for (var i = 0; i < keysLength; ++i) {
                if (isBitEnabledAt(seenBits, i)) {
                    continue;
                }
                var key = keys[i];
                var value = input[key];
                var result = _valueShapes[i]['_apply'](value, options);
                if (result === null) {
                    continue;
                }
                if (isArray(result)) {
                    unshiftPath(result, key);
                    if (!options.verbose) {
                        return result;
                    }
                    issues = concatIssues(issues, result);
                    continue;
                }
                if ((_unsafe || issues === null) && !isEqual(value, result.value)) {
                    if (input === output) {
                        output = cloneObjectEnumerableKeys(input);
                    }
                    setKeyValue(output, key, result.value);
                }
            }
        }
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && input !== output) {
            return ok(output);
        }
        return issues;
    };
    return ObjectShape;
}(Shape));

/**
 * The shape of a value wrapped in a `Promise` instance.
 *
 * @template S The shape of the resolved value.
 */
var PromiseShape = /** @class */ (function (_super) {
    __extends(PromiseShape, _super);
    /**
     * Creates a new {@linkcode PromiseShape} instance.
     *
     * @param shape The shape of the resolved value.
     * @param options The type constraint options or the type issue message.
     * @template S The shape of the resolved value.
     */
    function PromiseShape(shape, options) {
        var _this = _super.call(this) || this;
        _this.shape = shape;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_PROMISE_TYPE, options, TYPE_PROMISE);
        return _this;
    }
    PromiseShape.prototype._requiresAsync = function () {
        return true;
    };
    PromiseShape.prototype._getInputTypes = function () {
        return [TYPE_OBJECT];
    };
    PromiseShape.prototype._apply = function (input, options) {
        throw new Error(ERROR_REQUIRES_ASYNC);
    };
    PromiseShape.prototype._applyAsync = function (input, options) {
        var _this = this;
        var output = input;
        if (!(output instanceof Promise)) {
            if (!(options.coerced || this._coerced)) {
                return Promise.resolve(this._typeIssueFactory(input, options));
            }
            output = Promise.resolve(input);
        }
        var _applyChecks = this._applyChecks;
        var inputValue;
        var outputValue;
        return output
            .then(function (value) {
            inputValue = outputValue = value;
            return _this.shape['_applyAsync'](value, options);
        })
            .then(function (result) {
            var issues;
            if (result !== null) {
                if (isArray(result)) {
                    return result;
                }
                outputValue = result.value;
                if (!isEqual(inputValue, outputValue)) {
                    output = Promise.resolve(outputValue);
                }
            }
            if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
                return ok(output);
            }
            return issues;
        });
    };
    return PromiseShape;
}(CoercibleShape));

/**
 * The shape that describes an object with string keys and values that conform the given shape.
 *
 * @template K The key shape.
 * @template V The value shape.
 */
var RecordShape = /** @class */ (function (_super) {
    __extends(RecordShape, _super);
    /**
     * Creates a new {@linkcode RecordShape} instance.
     *
     * @param keyShape The key shape, or `null` if keys should be preserved intact.
     * @param valueShape The value shape.
     * @param options The type constraint options or an issue message.
     * @template K The key shape.
     * @template V The value shape.
     */
    function RecordShape(
    /**
     * The key shape or `null` if keys are preserved intact.
     */
    keyShape, 
    /**
     * The value shape.
     */
    valueShape, options) {
        var _this = _super.call(this) || this;
        _this.keyShape = keyShape;
        _this.valueShape = valueShape;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_OBJECT_TYPE, options, TYPE_OBJECT);
        return _this;
    }
    RecordShape.prototype.at = function (key) {
        return typeof key === 'string' || typeof key === 'number' ? this.valueShape : null;
    };
    RecordShape.prototype._requiresAsync = function () {
        return (this.keyShape !== null && this.keyShape.async) || this.valueShape.async;
    };
    RecordShape.prototype._getInputTypes = function () {
        return [TYPE_OBJECT];
    };
    RecordShape.prototype._apply = function (input, options) {
        if (!isObjectLike(input)) {
            return this._typeIssueFactory(input, options);
        }
        var _a = this, keyShape = _a.keyShape, valueShape = _a.valueShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var keyCount = 0;
        var issues = null;
        var output = input;
        for (var key in input) {
            var value = input[key];
            var outputKey = key;
            var outputValue = value;
            if (keyShape !== null) {
                var keyResult = keyShape['_apply'](key, options);
                if (keyResult !== null) {
                    if (isArray(keyResult)) {
                        unshiftPath(keyResult, key);
                        if (!options.verbose) {
                            return keyResult;
                        }
                        issues = concatIssues(issues, keyResult);
                    }
                    else {
                        outputKey = keyResult.value;
                    }
                }
            }
            var valueResult = valueShape['_apply'](value, options);
            if (valueResult !== null) {
                if (isArray(valueResult)) {
                    unshiftPath(valueResult, key);
                    if (!options.verbose) {
                        return valueResult;
                    }
                    issues = concatIssues(issues, valueResult);
                }
                else {
                    outputValue = valueResult.value;
                }
            }
            if ((_unsafe || issues === null) && (key !== outputKey || !isEqual(value, outputValue))) {
                if (input === output) {
                    output = cloneObjectEnumerableKeys(input, keyCount);
                }
                setKeyValue(output, outputKey, outputValue);
            }
        }
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && input !== output) {
            return ok(output);
        }
        return issues;
    };
    RecordShape.prototype._applyAsync = function (input, options) {
        var _this = this;
        return new Promise(function (resolve) {
            if (!isObjectLike(input)) {
                resolve(_this._typeIssueFactory(input, options));
                return;
            }
            var _a = _this, keyShape = _a.keyShape, valueShape = _a.valueShape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
            var promises = [];
            for (var key in input) {
                var value = input[key];
                promises.push(key, keyShape !== null ? keyShape['_applyAsync'](key, options) : null, valueShape['_applyAsync'](value, options));
            }
            resolve(Promise.all(promises).then(function (results) {
                var resultsLength = results.length;
                var keyCount = 0;
                var issues = null;
                var output = input;
                for (var i = 0; i < resultsLength; i += 3) {
                    var key = results[i];
                    var value = input[key];
                    var outputKey = key;
                    var outputValue = value;
                    var keyResult = results[i + 1];
                    var valueResult = results[i + 2];
                    if (keyResult !== null) {
                        if (isArray(keyResult)) {
                            unshiftPath(keyResult, key);
                            if (!options.verbose) {
                                return keyResult;
                            }
                            issues = concatIssues(issues, keyResult);
                        }
                        else {
                            outputKey = keyResult.value;
                        }
                    }
                    if (valueResult !== null) {
                        if (isArray(valueResult)) {
                            unshiftPath(valueResult, key);
                            if (!options.verbose) {
                                return valueResult;
                            }
                            issues = concatIssues(issues, valueResult);
                        }
                        else {
                            outputValue = valueResult.value;
                        }
                    }
                    if ((_unsafe || issues === null) && (key !== outputKey || !isEqual(value, outputValue))) {
                        if (input === output) {
                            output = cloneObjectEnumerableKeys(input, keyCount);
                        }
                        output[outputKey] = outputValue;
                    }
                }
                if (_applyChecks !== null && (_unsafe || issues === null)) {
                    issues = _applyChecks(output, issues, options);
                }
                if (issues === null && input !== output) {
                    return ok(output);
                }
                return issues;
            }));
        });
    };
    return RecordShape;
}(Shape));

/**
 * The shape of a `Set` instance.
 *
 * @template S The value shape.
 */
var SetShape = /** @class */ (function (_super) {
    __extends(SetShape, _super);
    /**
     * Creates a new {@linkcode SetShape} instance.
     *
     * @param shape The value shape
     * @param options The type constraint options or the type issue message.
     * @template S The value shape.
     */
    function SetShape(
    /**
     * The value shape
     */
    shape, options) {
        var _this = _super.call(this) || this;
        _this.shape = shape;
        _this._options = options;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_SET_TYPE, options, TYPE_SET);
        return _this;
    }
    SetShape.prototype.at = function (key) {
        return toArrayIndex(key) === -1 ? null : this.shape;
    };
    /**
     * Constrains the set size.
     *
     * @param size The minimum set size.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    SetShape.prototype.size = function (size, options) {
        return this.min(size, options).max(size, options);
    };
    /**
     * Constrains the minimum set size.
     *
     * @param size The minimum set size.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    SetShape.prototype.min = function (size, options) {
        var issueFactory = createIssueFactory(CODE_SET_MIN, MESSAGE_SET_MIN, options, size);
        return setCheck(this, CODE_SET_MIN, options, size, function (input, options) {
            if (input.size < size) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the maximum set size.
     *
     * @param size The maximum set size.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    SetShape.prototype.max = function (size, options) {
        var issueFactory = createIssueFactory(CODE_SET_MAX, MESSAGE_SET_MAX, options, size);
        return setCheck(this, CODE_SET_MAX, options, size, function (input, options) {
            if (input.size > size) {
                return issueFactory(input, options);
            }
        });
    };
    SetShape.prototype._requiresAsync = function () {
        return this.shape.async;
    };
    SetShape.prototype._getInputTypes = function () {
        if (this._coerced) {
            return this.shape['_getInputTypes']().concat(TYPE_OBJECT, TYPE_ARRAY);
        }
        else {
            return [TYPE_OBJECT];
        }
    };
    SetShape.prototype._apply = function (input, options) {
        var changed = false;
        var values;
        var issues = null;
        if (
        // Not a Set
        !(input instanceof Set && (values = Array.from(input.values()))) &&
            // No coercion or not coercible
            (!(options.coerced || this._coerced) || !(changed = (values = this._coerceValues(input)) !== null))) {
            return this._typeIssueFactory(input, options);
        }
        var _a = this, shape = _a.shape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
        var valuesLength = values.length;
        for (var i = 0; i < valuesLength; ++i) {
            var value = values[i];
            var result = shape['_apply'](value, options);
            if (result === null) {
                continue;
            }
            if (isArray(result)) {
                unshiftPath(result, i);
                if (!options.verbose) {
                    return result;
                }
                issues = concatIssues(issues, result);
                continue;
            }
            changed = !isEqual(value, (values[i] = result.value));
        }
        var output = changed ? new Set(values) : input;
        if (_applyChecks !== null && (_unsafe || issues === null)) {
            issues = _applyChecks(output, issues, options);
        }
        if (issues === null && changed) {
            return ok(output);
        }
        return issues;
    };
    SetShape.prototype._applyAsync = function (input, options) {
        var _this = this;
        return new Promise(function (resolve) {
            var changed = false;
            var values;
            if (
            // Not a Set
            !(input instanceof Set && (values = Array.from(input.values()))) &&
                // No coercion or not coercible
                (!(options.coerced || _this._coerced) || !(changed = (values = _this._coerceValues(input)) !== null))) {
                resolve(_this._typeIssueFactory(input, options));
                return;
            }
            var _a = _this, shape = _a.shape, _applyChecks = _a._applyChecks, _unsafe = _a._unsafe;
            var valuesLength = values.length;
            var promises = [];
            for (var i = 0; i < valuesLength; ++i) {
                promises.push(shape['_applyAsync'](values[i], options));
            }
            resolve(Promise.all(promises).then(function (results) {
                var resultsLength = results.length;
                var issues = null;
                for (var i = 0; i < resultsLength; ++i) {
                    var result = results[i];
                    if (result === null) {
                        continue;
                    }
                    if (isArray(result)) {
                        unshiftPath(result, i);
                        if (!options.verbose) {
                            return result;
                        }
                        issues = concatIssues(issues, result);
                        continue;
                    }
                    changed = !isEqual(values[i], (values[i] = result.value));
                }
                var output = changed ? new Set(values) : input;
                if (_applyChecks !== null && (_unsafe || issues === null)) {
                    issues = _applyChecks(output, issues, options);
                }
                if (issues === null && changed) {
                    return ok(output);
                }
                return issues;
            }));
        });
    };
    /**
     * Coerces value to a list of `Set` values, or returns `null` if coercion isn't possible.
     *
     * @param value The non-`Set` value to coerce.
     */
    SetShape.prototype._coerceValues = function (value) {
        if (isArray(value)) {
            return value;
        }
        if (isIterable(value)) {
            return Array.from(value);
        }
        return [value];
    };
    return SetShape;
}(CoercibleShape));

/**
 * The shape that constrains the input as a string.
 */
var StringShape = /** @class */ (function (_super) {
    __extends(StringShape, _super);
    /**
     * Creates a new {@linkcode StringShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function StringShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_STRING_TYPE, options, TYPE_STRING);
        return _this;
    }
    /**
     * The shortcut to apply both {@linkcode min} and {@linkcode max} constraints.
     *
     * @param length The exact length a string must have.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    StringShape.prototype.length = function (length, options) {
        return this.min(length, options).max(length, options);
    };
    /**
     * Constrains the string length to be greater than or equal to the length.
     *
     * @param length The minimum string length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    StringShape.prototype.min = function (length, options) {
        var issueFactory = createIssueFactory(CODE_STRING_MIN, MESSAGE_STRING_MIN, options, length);
        return setCheck(this, CODE_STRING_MIN, options, length, function (input, options) {
            if (input.length < length) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the string length to be less than or equal to the length.
     *
     * @param length The maximum string length.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    StringShape.prototype.max = function (length, options) {
        var issueFactory = createIssueFactory(CODE_STRING_MAX, MESSAGE_STRING_MAX, options, length);
        return setCheck(this, CODE_STRING_MAX, options, length, function (input, options) {
            if (input.length > length) {
                return issueFactory(input, options);
            }
        });
    };
    /**
     * Constrains the string to match a regexp.
     *
     * @param re The regular expression that the sting must conform.
     * @param options The constraint options or an issue message.
     * @returns The clone of the shape.
     */
    StringShape.prototype.regex = function (re, options) {
        var issueFactory = createIssueFactory(CODE_STRING_REGEX, MESSAGE_STRING_REGEX, options, re);
        return setCheck(this, CODE_STRING_REGEX, options, re, function (input, options) {
            re.lastIndex = 0;
            if (!re.test(input)) {
                return issueFactory(input, options);
            }
        });
    };
    StringShape.prototype._getInputTypes = function () {
        if (this._coerced) {
            return [TYPE_STRING, TYPE_NUMBER, TYPE_BOOLEAN, TYPE_BIGINT, TYPE_ARRAY, TYPE_UNDEFINED, TYPE_NULL];
        }
        else {
            return [TYPE_STRING];
        }
    };
    StringShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var output = input;
        var issues = null;
        var changed = false;
        if (typeof output !== 'string' &&
            (!(changed = options.coerced || this._coerced) || (output = this._coerce(input)) === null)) {
            return this._typeIssueFactory(input, options);
        }
        if ((_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) && changed) {
            return ok(output);
        }
        return issues;
    };
    /**
     * Coerces value to a string or returns `null` if coercion isn't possible.
     *
     * @param value The non-string value to coerce.
     */
    StringShape.prototype._coerce = function (value) {
        if (isArray(value) && value.length === 1 && typeof (value = value[0]) === 'string') {
            return value;
        }
        if (value === null || value === undefined) {
            return '';
        }
        if (Number.isFinite(value) || typeof value === 'boolean' || typeof value === 'bigint') {
            return '' + value;
        }
        if (isDate(value)) {
            return value.toISOString();
        }
        return null;
    };
    return StringShape;
}(CoercibleShape));

/**
 * The shape of the arbitrary symbol.
 */
var SymbolShape = /** @class */ (function (_super) {
    __extends(SymbolShape, _super);
    /**
     * Creates a new {@linkcode SymbolShape} instance.
     *
     * @param options The type constraint options or the type issue message.
     */
    function SymbolShape(options) {
        var _this = _super.call(this) || this;
        _this._typeIssueFactory = createIssueFactory(CODE_TYPE, MESSAGE_SYMBOL_TYPE, options, TYPE_SYMBOL);
        return _this;
    }
    SymbolShape.prototype._getInputTypes = function () {
        return [TYPE_SYMBOL];
    };
    SymbolShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        if (typeof input !== 'symbol') {
            return this._typeIssueFactory(input, options);
        }
        if (_applyChecks !== null) {
            return _applyChecks(input, null, options);
        }
        return null;
    };
    return SymbolShape;
}(Shape));

/**
 * The shape that requires an input to conform at least one of shapes.
 *
 * @template U The list of shapes that comprise a union.
 */
var UnionShape = /** @class */ (function (_super) {
    __extends(UnionShape, _super);
    /**
     * Creates a new {@linkcode UnionShape} instance.
     *
     * @param shapes The list of shapes that comprise a union.
     * @param options The union constraint options or an issue message.
     * @template U The list of shapes that comprise a union.
     */
    function UnionShape(
    /**
     * The list of shapes that comprise a union.
     */
    shapes, options) {
        var _this = _super.call(this) || this;
        _this.shapes = shapes;
        _this._options = options;
        _this._typeIssueFactory = createIssueFactory(CODE_UNION, MESSAGE_UNION, options);
        return _this;
    }
    Object.defineProperty(UnionShape.prototype, "_lookup", {
        get: function () {
            var cb = createDiscriminatorLookupCallback(this.shapes) || createValueTypeLookupCallback(this.shapes);
            Object.defineProperty(this, '_lookup', { value: cb });
            return cb;
        },
        enumerable: false,
        configurable: true
    });
    UnionShape.prototype.at = function (key) {
        var valueShapes = [];
        for (var _i = 0, _a = this.shapes; _i < _a.length; _i++) {
            var shape = _a[_i];
            var valueShape = shape.at(key);
            if (valueShape !== null) {
                valueShapes.push(valueShape);
            }
        }
        if (valueShapes.length === 0) {
            return null;
        }
        if (valueShapes.length === 1) {
            return valueShapes[0];
        }
        return new UnionShape(valueShapes);
    };
    UnionShape.prototype._requiresAsync = function () {
        return isAsyncShapes(this.shapes);
    };
    UnionShape.prototype._getInputTypes = function () {
        var inputTypes = [];
        for (var _i = 0, _a = this.shapes; _i < _a.length; _i++) {
            var shape = _a[_i];
            inputTypes.push.apply(inputTypes, shape['_getInputTypes']());
        }
        return unique(inputTypes);
    };
    UnionShape.prototype._getInputValues = function () {
        var inputValues = [];
        for (var _i = 0, _a = this.shapes; _i < _a.length; _i++) {
            var shape = _a[_i];
            var values = shape['_getInputValues']();
            if (values.length === 0) {
                return [];
            }
            inputValues.push.apply(inputValues, values);
        }
        return inputValues;
    };
    UnionShape.prototype._apply = function (input, options) {
        var _applyChecks = this._applyChecks;
        var result = null;
        var issues = null;
        var output = input;
        var index;
        var shapes = this._lookup(input);
        var shapesLength = shapes.length;
        for (index = 0; index < shapesLength; ++index) {
            result = shapes[index]['_apply'](input, options);
            if (result === null) {
                break;
            }
            if (isArray(result)) {
                issues = concatIssues(issues, result);
                continue;
            }
            output = result.value;
            break;
        }
        if (index === shapesLength) {
            return issues !== null ? issues : this._typeIssueFactory(input, options, this._getInputTypes());
        }
        if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
            return result;
        }
        return issues;
    };
    UnionShape.prototype._applyAsync = function (input, options) {
        var _applyChecks = this._applyChecks;
        var shapes = this._lookup(input);
        var shapesLength = shapes.length;
        if (shapesLength === 0) {
            return Promise.resolve(this._typeIssueFactory(input, options, this._getInputTypes()));
        }
        var issues = null;
        var index = 0;
        var nextShape = function () {
            return shapes[index]['_applyAsync'](input, options).then(function (result) {
                ++index;
                var output = input;
                if (result !== null) {
                    if (isArray(result)) {
                        issues = concatIssues(issues, result);
                        if (index === shapesLength) {
                            return issues;
                        }
                        return nextShape();
                    }
                    output = result.value;
                }
                if (_applyChecks === null || (issues = _applyChecks(output, null, options)) === null) {
                    return result;
                }
                return issues;
            });
        };
        return nextShape();
    };
    return UnionShape;
}(Shape));
/**
 * Creates a lookup that finds a shape using an input value type.
 */
function createValueTypeLookupCallback(shapes) {
    var buckets = {
        object: [],
        array: [],
        function: [],
        string: [],
        symbol: [],
        number: [],
        bigint: [],
        boolean: [],
        date: [],
        null: [],
        undefined: [],
    };
    var bucketTypes = Object.keys(buckets);
    for (var _i = 0, _a = unique(shapes); _i < _a.length; _i++) {
        var shape = _a[_i];
        var inputTypes = shape['_getInputTypes']();
        for (var _b = 0, _c = inputTypes.includes(TYPE_ANY) ? bucketTypes : unique(inputTypes); _b < _c.length; _b++) {
            var type = _c[_b];
            if (type === TYPE_ANY || type === TYPE_NEVER) {
                continue;
            }
            buckets[type].push(shape);
        }
    }
    return function (input) { return buckets[getValueType(input)]; };
}
/**
 * Creates a lookup that uses a discriminator property, or returns `null` if discriminator cannot be detected.
 */
function createDiscriminatorLookupCallback(shapes) {
    var shapesLength = shapes.length;
    if (shapesLength <= 1 || !shapes.every(isObjectShape)) {
        return null;
    }
    var discriminator = getDiscriminator(shapes);
    if (discriminator === null) {
        return null;
    }
    var key = discriminator.key, valuesForShape = discriminator.valuesForShape;
    var shapeArrays = shapes.map(function (shape) { return [shape]; });
    var noShapesArray = [];
    if (valuesForShape.every(function (values) { return values.length === 1 && values[0] === values[0]; })) {
        var values_1 = valuesForShape.map(function (values) { return values[0]; });
        return function (input) {
            if (!isObjectLike(input)) {
                return noShapesArray;
            }
            var index = values_1.indexOf(input[key]);
            if (index === -1) {
                return noShapesArray;
            }
            return shapeArrays[index];
        };
    }
    return function (input) {
        if (!isObjectLike(input)) {
            return noShapesArray;
        }
        var value = input[key];
        for (var i = 0; i < shapesLength; ++i) {
            if (valuesForShape[i].includes(value)) {
                return shapeArrays[i];
            }
        }
        return noShapesArray;
    };
}
function getDiscriminator(shapes) {
    var keys = shapes[0].keys.slice(0);
    var valuesForShapeForKey = [];
    for (var i = 0; i < shapes.length && keys.length !== 0; ++i) {
        var shapeValues = [];
        var _a = shapes[i], shapeKeys = _a.keys, shapeShapes = _a.shapes;
        valuesForShapeForKey[i] = shapeValues;
        for (var j = 0; j < keys.length; ++j) {
            var key = keys[j];
            if (shapeKeys.includes(key)) {
                var keyValues = shapeShapes[key]['_getInputValues']();
                if (keyValues.length !== 0) {
                    var duplicated = false;
                    // Ensure that this set of values wasn't seen for this key in other shapes
                    duplicateLookup: for (var k = 0; k < i; ++k) {
                        for (var _i = 0, _b = valuesForShapeForKey[k][j]; _i < _b.length; _i++) {
                            var value = _b[_i];
                            if (keyValues.includes(value)) {
                                duplicated = true;
                                break duplicateLookup;
                            }
                        }
                    }
                    if (!duplicated) {
                        shapeValues.push(keyValues);
                        continue;
                    }
                }
            }
            keys.splice(j, 1);
            for (var k = 0; k < i; ++k) {
                valuesForShapeForKey[k].splice(j, 1);
            }
            --j;
        }
    }
    if (keys.length === 0) {
        return null;
    }
    return {
        key: keys[0],
        valuesForShape: valuesForShapeForKey.map(function (values) { return values[0]; }),
    };
}
function isObjectShape(shape) {
    return shape instanceof ObjectShape;
}

function any(cb, options) {
    var shape = new Shape();
    return cb != null ? shape.refine(cb, options) : shape;
}

function array(shape, options) {
    if (shape == null || shape instanceof Shape) {
        return new ArrayShape(null, shape || null, options);
    }
    else {
        return new ArrayShape(null, null, shape);
    }
}

/**
 * Creates the bigint shape.
 *
 * @param options The constraint options or an issue message.
 */
function bigint(options) {
    return new BigIntShape(options);
}

/**
 * Creates the boolean shape.
 *
 * @param options The constraint options or an issue message.
 */
function boolean(options) {
    return new BooleanShape(options);
}

/**
 * Creates the constant value shape.
 *
 * @param value The value to which the input must be strictly equal.
 * @param options The constraint options or an issue message.
 * @template T The value type.
 */
function const_(value, options) {
    return new ConstShape(value, options);
}

/**
 * Creates the `Date` shape.
 *
 * @param options The constraint options or an issue message.
 */
function date(options) {
    return new DateShape(options);
}

function enum_(source, options) {
    return new EnumShape(source, options);
}

/**
 * Creates the finite number shape.
 *
 * @param options The constraint options or an issue message.
 */
function finite(options) {
    return new NumberShape(options).finite();
}

/**
 * Creates the class instance shape.
 *
 * @param ctor The instance constructor.
 * @param options The constraint options or an issue message.
 * @template C The instance constructor.
 */
function instanceOf(ctor, options) {
    return new InstanceShape(ctor, options);
}

/**
 * Creates an intersection shape that tries to parse the input with all provided shapes.
 *
 * @param shapes The list of shapes.
 * @param options The constraint options or an issue message.
 * @template U The tuple of intersected shapes.
 */
function intersection(shapes, options) {
    return new IntersectionShape(shapes, options);
}

/**
 * Creates the shape that parses JSON strings.
 *
 * @param options The constraint options or an issue message.
 */
function json(options) {
    return new JSONShape(options);
}

/**
 * Creates the shape that resolves the underlying shape on-demand.
 *
 * @param shapeProvider The provider that returns the base shape.
 */
function lazy(shapeProvider) {
    return new LazyShape(shapeProvider);
}

/**
 * Creates the `Map` instance shape.
 *
 * @param keyShape The key shape.
 * @param valueShape The value shape.
 * @param options The type constraint options or an issue message.
 * @template K The key shape.
 * @template V The value shape.
 */
function map(keyShape, valueShape, options) {
    return new MapShape(keyShape, valueShape, options);
}

/**
 * Creates the shape that requires an input to equal to `NaN`.
 *
 * @param options The constraint options or an issue message.
 */
function nan(options) {
    return new ConstShape(NaN, options);
}

/**
 * Creates the integer shape.
 *
 * @param options The constraint options or an issue message.
 */
function integer(options) {
    return new NumberShape(options).integer(options);
}

/**
 * Creates the shape that always raises an issue.
 *
 * @param options The constraint options or an issue message.
 */
function never(options) {
    return new NeverShape(options);
}

/**
 * Creates the shape that requires an input to equal to `null`.
 *
 * @param options The constraint options or an issue message.
 */
function null_(options) {
    return new ConstShape(null, options);
}

/**
 * Creates the number shape.
 *
 * @param options The constraint options or an issue message.
 */
function number(options) {
    return new NumberShape(options);
}

/**
 * Creates the object shape.
 *
 * @param shapes The mapping from an object key to a corresponding shape.
 * @param options The constraint options or an issue message.
 */
function object(shapes, options) {
    return new ObjectShape(shapes, null, options);
}

/**
 * Creates the `Promise` instance shape.
 *
 * @param shape The shape of the resolved value.
 * @param options The constraint options or an issue message.
 * @template S The shape of the resolved value.
 */
function promise(shape, options) {
    return new PromiseShape(shape, options);
}

function record(keyShape, valueShape, options) {
    if (valueShape instanceof Shape) {
        return new RecordShape(keyShape, valueShape, options);
    }
    else {
        return new RecordShape(null, keyShape, options);
    }
}

/**
 * Creates the `Set` instance shape.
 *
 * @param shape The value shape
 * @param options The constraint options or an issue message.
 * @template S The value shape.
 */
function set(shape, options) {
    return new SetShape(shape, options);
}

/**
 * Creates the string shape.
 *
 * @param options The constraint options or an issue message.
 */
function string(options) {
    return new StringShape(options);
}

/**
 * Creates the symbol shape.
 *
 * @param options The constraint options or an issue message.
 */
function symbol(options) {
    return new SymbolShape(options);
}

/**
 * Creates the shape that synchronously transforms the input value.
 *
 * @param cb The callback that processes the input value.
 * @template I The input value.
 * @template O The output value.
 */
function transform(cb) {
    return new Shape().transform(cb);
}
/**
 * Creates the shape that asynchronously transforms the input value.
 *
 * @param cb The callback that processes the input value.
 * @template I The input value.
 * @template O The output value.
 */
function transformAsync(cb) {
    return new Shape().transformAsync(cb);
}

function tuple(shapes, restShape, options) {
    if (restShape == null || restShape instanceof Shape) {
        return new ArrayShape(shapes, restShape || null, options);
    }
    else {
        return new ArrayShape(shapes, null, restShape);
    }
}

/**
 * Creates the shape that requires an input to equal to `undefined`.
 *
 * @param options The constraint options or an issue message.
 */
function undefined_(options) {
    return new ConstShape(undefined, options);
}

/**
 * Creates a union shape that tries to parse the input with one of the provided shapes.
 *
 * @param shapes The list of shapes to try.
 * @param options The constraint options or an issue message.
 * @template U The tuple of united shapes.
 */
function union(shapes, options) {
    return new UnionShape(shapes, options);
}

/**
 * Creates the unconstrained shape with unknown value.
 */
function unknown() {
    return new Shape();
}

/**
 * Creates a shape that requires an input to be `undefined` at runtime and typed as `void`.
 *
 * @param options The constraint options or an issue message.
 */
function void_(options) {
    return new ConstShape(undefined, options);
}

function guard(shape, callback) {
    if (isArray(shape)) {
        shape = new ArrayShape(shape, null);
    }
    if (!(shape instanceof ArrayShape)) {
        shape = new ArrayShape([shape], null);
    }
    if (shape.async) {
        throw new Error(ERROR_ASYNC_ARGUMENTS);
    }
    var factory = function (callback) {
        return function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return callback.apply(this, shape.parse(args));
        };
    };
    return callback !== undefined ? factory(callback) : factory;
}
function guardAsync(shape, callback) {
    if (isArray(shape)) {
        shape = new ArrayShape(shape, null);
    }
    if (!(shape instanceof ArrayShape)) {
        shape = new ArrayShape([shape], null);
    }
    var factory;
    if (shape.async) {
        factory = function (callback) {
            return function () {
                var _this = this;
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                return shape.parseAsync(args).then(function (args) { return callback.apply(_this, shape.parse(args)); });
            };
        };
    }
    else {
        factory = function (callback) {
            return function () {
                var _this = this;
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                return new Promise(function (resolve) { return resolve(callback.apply(_this, shape.parse(args))); });
            };
        };
    }
    return callback !== undefined ? factory(callback) : factory;
}

export { ArrayShape, BigIntShape, BooleanShape, CatchShape, CoercibleShape, ConstShape, DateShape, EnumShape, ExcludeShape, InstanceShape, IntersectionShape, JSONShape, LazyShape, MapShape, NeverShape, NumberShape, ObjectShape, PipeShape, PromiseShape, RecordShape, ReplaceShape, SetShape, Shape, StringShape, SymbolShape, TransformShape, UnionShape, ValidationError, intersection as and, any, array, bigint, boolean as bool, boolean, const_ as const, createIssueFactory, date, enum_ as enum, finite, guard, guardAsync, instanceOf, integer as int, integer, intersection, json, lazy, map, nan, never, null_ as null, number, object, union as or, promise, record, set, string, symbol, transform, transformAsync, tuple, undefined_ as undefined, union, unknown, void_ as void };
